from uret.transformers import SubTransformer
from uret.transformers.binary.subtransformers import binary_to_bytez

import random
from copy import deepcopy


class SectionAdd(SubTransformer):
    name = "SectionAdd"

    def __init__(self, length=[6, 12], types=None, transformation_range=None, seed=None, subtransformer_index=None):
        """
        Initialize a `SectionAdd` object. This object adds a new section of a certain type to the binary
        :param length: The maximum or  [lower, upper] bound in bytes of section length.
        :param types: A list of section type names.
        :param seed: Random seed value
        :param subtransformer_index: An int indicating where in the transformation record the subtransformer is defined. This is used to enforce "max # of transformations for this subtransformer" constraint.
        """

        if isinstance(length, list):
            self.min_val = length[0]
            self.max_val = length[1]
        else:
            self.min_val = 1
            self.max_val = length[1]

        if types is None:
            self.types = [
                lief.PE.SECTION_TYPES.BSS,
                lief.PE.SECTION_TYPES.DATA,
                lief.PE.SECTION_TYPES.EXPORT,
                lief.PE.SECTION_TYPES.IDATA,
                lief.PE.SECTION_TYPES.RELOCATION,
                lief.PE.SECTION_TYPES.RESOURCE,
                lief.PE.SECTION_TYPES.TEXT,
                lief.PE.SECTION_TYPES.TLS_,
                lief.PE.SECTION_TYPES.UNKNOWN,
            ]
        else:
            import lief # lgtm [py/repeated-import]
            type_mapper = {
                "bss": lief.PE.SECTION_TYPES.BSS,
                "data": lief.PE.SECTION_TYPES.DATA,
                "export": lief.PE.SECTION_TYPES.EXPORT,
                "idata": lief.PE.SECTION_TYPES.IDATA,
                "relocation": lief.PE.SECTION_TYPES.RELOCATION,
                "resource": lief.PE.SECTION_TYPES.RESOURCE,
                "text": lief.PE.SECTION_TYPES.TEXT,
                "tls": lief.PE.SECTION_TYPES.TLS_,
                "unknown": lief.PE.SECTION_TYPES.UNKNOWN,
            }

            self.types = []
            for t in types:
                if t in type_mapper.keys():
                    self.types.append(type_mapper[t])
                else:
                    print(t + " is not a recognized type. Skipping")

        if len(self.types) == 0:
            raise ValueError("type list is empty")
        self.seed = seed
        self.subtransformer_index = subtransformer_index

        super(SectionAdd, self).__init__()

    def transform(self, x, transformation_record, transformation_value=None):
        """
        Add a new section to the binary
        :param x: Input Value
        :param transformation_record: Record of changes
        :param transformation_value: type of section to add
        :return: A transformed input and modified transformation record
        """
        import lief # lgtm [py/repeated-import]

        random.seed(self.seed)
        binary = lief.PE.parse(list(x))

        section_name = "".join(chr(random.randrange(ord("."), ord("z"))) for _ in range(6))
        new_section = lief.PE.Section(section_name)

        if transformation_value is None:
            transformation_value = random.choice(self.types)

        # fill section with random stuff
        upper = random.randrange(256)
        content_length = random.randint(self.min_val, self.max_val)
        new_section.content = [random.randint(0, upper) for _ in range(content_length)]
        new_section.virtual_address = max([s.virtual_address + s.size for s in binary.sections])
        binary.add_section(new_section, transformation_value)

        new_x = binary_to_bytez(binary)

        # update the transformation record
        transformation_record["prev_state"] = deepcopy(transformation_record["current_state"])
        if self.subtransformer_index is not None:
            transformation_record["current_state"]["actions_taken"][self.subtransformer_index] += 1
        else:
            transformation_record["current_state"]["actions_taken"] += 1

        transformation_record["current_state"]["value"] = new_x

        return new_x, transformation_record

    def get_action_list(self):
        """
        Return a list of section types
        """
        return self.types

    def is_possible(self, x, transformation_value):
        """
        Determines if the action specified by transformation_value can be performed on the current input.
        :param x: Input value
        :param transformation_value: The definition of the transformation to apply

        :return: If action is possible and the transformation value. The effect on the record is not used
        """
        return True, [transformation_value], None

    def get_possible(self, x):
        """
        Determines what actions can be performed on the input.
        :param x: input_value

        :return: A list of transformation values
        """
        return self.types



import numpy as np


def feature_sum(input_state, indices, dependency_indices):
    a, b, c = dependency_indices
    input_state[indices] = input_state[a] + input_state[b] - input_state[c]
    return input_state


def normalize(input_state, indices, dependency_indices):
    total_sum = np.sum([input_state[i] for i in dependency_indices])

    if total_sum != 0:
        for i in indices:
            input_state[i] = input_state[i] / total_sum

    return input_state



from uret.transformers import Transformer
import numpy as np
import random


class StringTransformer(Transformer):
    name = "StringTransformer"

    # Two attributes used by simulated annealing to track transformer properties
    transform_initial = True  # If True, then inputs to transform() should be based on the initial state of the sample
    transform_repeat = (
        True  # If True, then transform() should be called multiple times during one annealing move() call
    )

    def __init__(self, subtransformer_args, input_constraints={}, input_processor=None):
        """
        Initialize a `StringTransformer` object.
        :param subtransformer_list: The functions to be used to transform the input of a certain data type.
        :param input_constraints: The constraints to enforce on the input after transformation. The possible enforcement values
        are:
            - max_actions: An int or float representing the maximum number of subtransformer actions allowed to be performed. If an int it represents an absolute constraint. If a float, it represents a relative constraint with respect to the original length of the string.
            - max_subtransformer_actions: A optional list of values with length equal to the number of subtransformers. Each element represents the max number of times the corresponding subtransformer can be called. If an int it represents an absolute constraint. If a float, it represents a relative constraint with respect to the original length of the string.
            - "eps": The maximum change in the length of the string". If the value is an int it represents an absolute constraint.
            If it is a float, it represents a relative constraint with respect to the original length of the string.
        :param input_processer: A function with two modes to control transforming input values where parts of the input should
        remain unchanged. If a single input is given, the function will return (the modifiable sections, unmodifiable sections).
        If two inputs are given, it is assume they are in the order (the modifiable sections, unmodifiable sections) and the
        function will fuse them together.
        """

        for key in input_constraints:
            if key not in ["max_actions", "max_subtransformer_actions", "eps"]:
                raise ValueError(key + " is an unrecognized constraint type")
            elif key == "max_subtransformer_actions":
                if isinstance(input_constraints[key], list):
                    if any(np.array(input_constraints[key])) <= 0:
                        raise ValueError(key + " cannot be <= 0")
                    if len(subtransformer_args) != len(input_constraints[key]):
                        raise ValueError(key + " must be the same length as the number of subtransformers if a list")
                    # This is so the subtransformer will know the output
                    # format of the transformation record
                    for i, args in enumerate(subtransformer_args):
                        if "init_args" not in args.keys():
                            args["init_args"] = {}
                        args["init_args"]["subtransformer_index"] = i
                elif input_constraints[key] <= 0:
                    raise ValueError(key + " cannot be <= 0")

            elif input_constraints[key] <= 0:
                raise ValueError(key + " cannot be 0 or negative")

        super(StringTransformer, self).__init__(subtransformer_args, input_constraints, input_processor)

    def init_transformation_record(self, x):
        """
        Define the structure of the transformation record and initialize it.
        :param x: The input value to be transformed

        :return: A record that records the previous and current string state so we can undo changes if needed. We only look 1
        state back because the prev state is guaranteed to be valid and it is the next state that must be verified
        """

        actions_taken = 0
        if "max_actions" in self.input_constraints:
            if self.input_constraints["max_actions"] < 1:
                max_actions = np.ceil(len(x) * self.input_constraints["max_actions"])
            else:
                max_actions = int(self.input_constraints["max_actions"])
        else:
            max_actions = None
            
        if "max_subtransformer_actions" in self.input_constraints:
                max_subtransformer_actions = []
                actions_taken = []
                for v in self.input_constraints["max_subtransformer_actions"]:
                    if v < 1:
                        max_subtransformer_actions.append(len(x) * v)
                    else:
                        max_subtransformer_actions.append(int(v))
                    actions_taken.append(0)
        else:
            max_subtransformer_actions = None

        if "eps" in self.input_constraints:
            if self.input_constraints["eps"] < 1:
                eps = np.ceil(len(x) * self.input_constraints["eps"])
            else:
                eps = int(self.input_constraints["eps"])
        else:
            eps = None

        return {
            "prev_state": {"actions_taken": actions_taken, "delta": 0, "value": x},
            "current_state": {"actions_taken": actions_taken, "delta": 0, "value": x},
            "max_subtransformer_actions": max_subtransformer_actions,
            "max_actions": max_actions,
            "eps": eps,
        }

    def _enforce_constraints(self, x, transformation_record):
        """
        Enforce predefined input constraints
        :param x: Input value to be modified
        :param transformation_record: A record tracking the transformations already applied to x.

        :return: A "clipped" input and corrected transformation record. If clipping is done, the function "rewinds" back to the
        previous state
        """

        clipped_x = x
        corrected_transformation_record = transformation_record

        if "max_actions" in self.input_constraints:
            if (isinstance(transformation_record["current_state"]["actions_taken"], list) and np.sum(transformation_record["current_state"]["actions_taken"]) > transformation_record["max_actions"]) or (isinstance(transformation_record["current_state"]["actions_taken"], int) and transformation_record["current_state"]["actions_taken"] > transformation_record["max_actions"]): 
                clipped_x = transformation_record["prev_state"]["value"]  # The string prior to transformation
                corrected_transformation_record["current_state"] = corrected_transformation_record["prev_state"].copy()

        if "max_subtransformer_actions" in self.input_constraints and any(
                np.array(transformation_record["current_state"]["actions_taken"])
                > np.array(transformation_record["max_subtransformer_actions"])
            ):
                clipped_x = corrected_transformation_record["prev_state"]["value"]  # The string prior to transformation
                corrected_transformation_record["current_state"] = corrected_transformation_record["prev_state"].copy()

        if "eps" in self.input_constraints:
            if abs(transformation_record["current_state"]["delta"]) > transformation_record["eps"]:
                clipped_x = transformation_record["prev_state"]["value"]  # The string prior to transformation
                corrected_transformation_record["current_state"] = corrected_transformation_record["prev_state"].copy()

        return clipped_x, corrected_transformation_record

    def _is_possible(self, x, transformation_record=None, transformation_value=None, *args, **kwargs):
        """
        Determines if the action specified by tranformation_value can be performed on the current input. If it can, it returns
        True and provides a list of the arguments to use with transform().
        :param x: Input value
        :param transformation_value: The definition of the transformation to apply and the parameters to use
        :return: True if possible, list of possible arguments
        """

        if transformation_value is None:
            return False, []  # Can't verify default actions

        # Check action constraint. This can be done without consulting the subtransformer
        if "max_subtransformer_actions" in self.input_constraints:
            actions_taken = transformation_record["current_state"]["actions_taken"]
            next_action = np.zeros_like(actions_taken)
            next_action[transformation_value[0]] += 1
            predicted_taken = np.array(actions_taken) + next_action
            if any(predicted_taken > np.array(transformation_record["max_subtransformer_actions"])) or ("max_actions" in self.input_constraints and np.sum(predicted_taken) > transformation_record["max_actions"]):
                return False, []
        elif "max_actions" in self.input_constraints and transformation_record["current_state"]["actions_taken"] + 1 > transformation_record["max_actions"]:
                return False, []

        # Ask the subtransformer if it is possible
        possible, action_args, transformation_effect = self.subtransformer_list[transformation_value[0]].is_possible(
            x, *transformation_value[1:]
        )

        if not possible:
            return False, []

        # Check eps (ie length) constraint
        if transformation_record is not None:
            if (
                "eps" in self.input_constraints
                and abs(transformation_record["current_state"]["delta"] + transformation_effect)
                > transformation_record["eps"]
            ):
                return False, []

        return True, action_args



import numpy as np

from uret.core.rankers import BruteForce, LookupTable, ExternalModel, Random
from uret.core.explorers import (
    BeamSearchGraphExplorer,
    GreedySearchGraphExplorer,
    SimulatedAnnealingSearchGraphExplorer,
)


def process_config_file(config_filepath, model, feature_extractor=None, input_processor_list=[]):
    """
    Processes a config file and returns the explorer
    config_filepath: Path to the config file to parse
    model: Model to be attacked with a function that returns the prediction vector
    feature_extractor: Converts input data into a form the model can ingest
    input_processor_dict: Mapping of function names to functions
    """
    import os.path
    import yaml
    import importlib

    with open(config_filepath, "r") as stream:
        config_params = yaml.safe_load(stream)

    transformer_params = config_params.get("transformer_params", None)
    ranker_params = config_params.get("ranker_params", None)
    explorer_params = config_params.get("explorer_params", None)
    dependency_params = config_params.get("dependency_params", None)

    if transformer_params is None:
        raise RuntimeError(f"transformer_params was not found in the config file")
    if ranker_params is None:
        raise RuntimeError(f"ranker_params was not found in the config file")
    if explorer_params is None:
        raise RuntimeError(f"explorer_params was not found in the config file")

    # 1.Create the Transformer List
    # Assume transformers for data types are in the uret.transformers directory
    transformer_list = []
    for tp in transformer_params:
        data_type = tp["data_type"]
        init_args = tp.get("init_args", {})

        if data_type.lower() == "category" or data_type.lower() == "number" or data_type.lower() == "string":
            transformer_path = "uret.transformers.basic"
        else:
            transformer_path = "uret.transformers." + data_type.lower()

        transformer_module = importlib.import_module(transformer_path)
        transformer = getattr(transformer_module, data_type.lower().capitalize() + "Transformer")
        input_processor_name = tp.get("input_processor_name", None)
        if input_processor_name is not None:
            found = False
            for ip in input_processor_list:
                if ip.__name__ == input_processor_name:
                    init_args["input_processor"] = ip
                    found = True
                    break
            if not found:
                raise ValueError(f"No entry for {input_processor_name} in input_processor_dict")

        if len(init_args.keys()) > 0:
            transformer = transformer(**init_args)
        else:
            transformer = transformer()

        feature_index = tp.get("feature_index", None)
        if (
            "init_args" in ranker_params
            and ranker_params["init_args"].get("multi_feature_input", False)
            and feature_index is None
        ):
            raise ValueError(
                f"Ranker indicates multi-feature input transformations, but the transformer doesn't specify which feature index it modifies"
            )

        transformer_list.append([transformer, feature_index])

    # 2.Create the Ranker
    if ranker_params["type"].replace("_", "").lower() in ["brute", "bruteforce"]:
        ranking_alg = BruteForce
    elif ranker_params["type"].replace("_", "").lower() in ["lookup", "lookuptable"]:
        ranking_alg = LookupTable
    elif ranker_params["type"].replace("_", "").lower() in ["external", "externalmodel"]:
        ranking_alg = ExternalModel
    elif ranker_params["type"].lower() == "random":
        ranking_alg = Random
    else:
        raise ValueError(f"{ranker_params['type']} is not a recognized ranker type")

    if "init_args" in ranker_params.keys():
        ranker = ranking_alg(transformer_list, **ranker_params["init_args"])
    else:
        ranker = transformer(transformer_list)

    # 3.Create the Explorer and return
    explorer_init_args = explorer_params.get("init_args", {})
    explorer_init_args["ranking_algorithm"] = ranker

    # Create model values
    predict_function_name = explorer_params.get("predict_function_name", "predict")
    if not hasattr(model, predict_function_name):
        raise ValueError(f"{predict_function_name} is not an attribute of model")
    explorer_init_args["model_predict"] = getattr(model, predict_function_name)
    explorer_init_args["feature_extractor"] = feature_extractor

    # Create dependencies
    if dependency_params is not None:
        dependencies = dependency_params.get("dependencies", [])
        if len(dependencies) > 0:
            dependency_path = (
                "utils.dependency_functions." + dependency_params.get("dependency_path", "default").split(".")[0]
            )
            dependency_module = importlib.import_module(dependency_path)
            explorer_init_args["dependencies"] = []
            for d in dependencies:
                d_func = getattr(dependency_module, d["name"])
                explorer_init_args["dependencies"].append([d_func, d["args"]])

    if explorer_params["type"].replace("_", "").lower() in ["beam", "beamsearch", "beamsearchgraphexplorer"]:
        explorer_alg = BeamSearchGraphExplorer
    elif explorer_params["type"].replace("_", "").lower() in ["greedy", "greedysearch", "greedysearchgraphexplorer"]:
        explorer_alg = GreedySearchGraphExplorer
    elif explorer_params["type"].replace("_", "").lower() in [
        "simanneal",
        "simulatedannealing",
        "simulatedannealingsearchgraphexplorer",
    ]:
        explorer_alg = SimulatedAnnealingSearchGraphExplorer
    else:
        raise ValueError(f"{explorer_params['type']} is not a recognized explorer type")

    return explorer_alg(**explorer_init_args)



from uret.core.rankers.ranking_algorithm import RankingAlgorithm


class ExternalModel(RankingAlgorithm):
    def __init__(self, transformer_list, multi_feature_input=False, is_trained=True, model=None):
        """
        :param transformer_list: List of available transformations to be used by the algorithm.
        :param multi_feature_input: A boolean flag indicating if each index in the transformer
            list defines a transformer(s) for a single data type in the input. If False, then it
            is assumed that the transformer list defines multiple transformers for a single input
        :param model: A pretrained model who takes in an input sample and outputs the estimated score for every possible action
        """
        super().__init__(transformer_list, multi_feature_input, is_trained=True)

        if model is None:
            raise ValueError("A model must be provided")
        self.model = model

        # Enumerate all possible actions
        self.action_list = []
        for transformer_index, [transformer, input_index] in enumerate(transformer_list):
            if self.multi_feature_input:
                self.action_list.append(
                    [[transformer_index, input_index], transformer, action] for action in transformer.get_action_list()
                )
            else:
                self.action_list.append([None, transformer, action] for action in transformer.get_action_list())

    def rank_edges(self, sample, scoring_function, score_input, dependencies=[], current_transformation_records=None):
        scores = self.model(sample)[0]
        return [self.action_list[i] + [None, None] + s for i, s in enumerate(scores)]



from uret.core.rankers.ranking_algorithm import RankingAlgorithm

import numpy as np
import random
import copy
import warnings


class Random(RankingAlgorithm):
    """
    This implementation tries a random transformation for given sample.
    """

    def __init__(self, transformer_list, multi_feature_input=False, is_trained=True, num_actions=1):
        """
        :param transformer_list: List of available transformations to be used by the algorithm.
        :param multi_feature_input: A boolean flag indicating if each index in the transformer
            list defines a transformer(s) for a single data type in the input. If False, then it
            is assumed that the transformer list defines multiple transformers for a single input
        :param is_trained: Indicates when _train has been called at least once. Always True
        :param num_actions: Number of random actions to sample
        """
        super().__init__(transformer_list, multi_feature_input, is_trained=True)

        if num_actions < 1:
            num_actions = 1
        self.num_actions = num_actions

    def rank_edges(
        self,
        sample,
        scoring_function,
        score_input,
        dependencies=[],
        current_transformation_records=None,
        transformer_index=None,
    ):
        """
        param transformer_index: Index of the transformer to use for this ranking. This is a parameter used by annealing to randomly select an action for a specific transformer
        """

        # Create transformation record
        if self.multi_feature_input and current_transformation_records is None:
            current_transformation_records = [None for _ in range(len(self.transformer_list))]

        sample_temp = copy.copy(sample)
        transformation_records_temp = copy.copy(current_transformation_records)

        # Get a random transformer from the list. Make sure there is at least one possible action to perform
        # Exception is when annealing provides the specific index to us
        possible_actions = []
        if transformer_index is None:
            skip_inds = []
            while len(possible_actions) == 0:
                if len(skip_inds) == len(self.transformer_list):
                    return []  # No actions are possible

                transformer_index = random.choice(list(set(np.arange(len(self.transformer_list))) - set(skip_inds)))
                transformer, input_index = self.transformer_list[transformer_index]
                if self.multi_feature_input:
                    possible_actions = transformer.get_possible(
                        sample_temp[input_index], transformation_record=transformation_records_temp[transformer_index]
                    )
                else:
                    possible_actions = transformer.get_possible(
                        sample_temp, transformation_record=transformation_records_temp
                    )

                skip_inds.append(transformer_index)
        else:
            transformer, input_index = self.transformer_list[transformer_index]
            if self.multi_feature_input:
                possible_actions = transformer.get_possible(
                    sample_temp[input_index], transformation_record=transformation_records_temp[transformer_index]
                )
            else:
                possible_actions = transformer.get_possible(
                    sample_temp, transformation_record=transformation_records_temp
                )

            if len(possible_actions) == 0:
                return []  # No actions are possible

        return_values = []  # This will contain the (sample_index, transformer, action args, transformed sample,
        # transformation_record of the transformed sample, score)

        # Pick a random action  to perform
        actions = random.sample(possible_actions, k=min(len(possible_actions), self.num_actions))

        for action in actions:
            sample_temp = copy.copy(sample)
            transformation_records_temp = copy.copy(current_transformation_records)

            if self.multi_feature_input:
                transformed_value, new_transformation_record = transformer.transform(
                    sample_temp[input_index],
                    transformation_record=transformation_records_temp[transformer_index],
                    transformation_value=action,
                )
                sample_temp[input_index] = transformed_value
                transformation_records_temp[transformer_index] = new_transformation_record
            else:
                transformed_value, new_transformation_record = transformer.transform(
                    sample_temp, transformation_record=transformation_records_temp, transformation_value=action
                )
                sample_temp = transformed_value
                transformation_records_temp = new_transformation_record

            sample_temp = self._enforce_dependencies(sample_temp, dependencies)
            score = scoring_function(sample_temp, score_input)

            if self.multi_feature_input:
                return_values.append(
                    (
                        [transformer_index, input_index],
                        transformer,
                        action,
                        sample_temp,
                        transformation_records_temp,
                        score,
                    )
                )
            else:
                return_values.append((None, transformer, action, sample_temp, transformation_records_temp, score))

        return return_values



from uret.transformers import Transformer


class CategoryTransformer(Transformer):
    name = "CategoryTransformer"

    # Two attributes used by simulated annealing to track transformer properties
    transform_initial = True  # If True, then inputs to transform() should be based on the initial state of the sample

    def __init__(
        self,
        subtransformer_args,
        input_constraints={},
        input_processor=None,
    ):
        """
        Initialize a `CategoryTransformer` object.
        :param subtransformer_list: The functions to be used to transform the input of a certain data type.
        :param input_constraints: Not used
        :param input_processor: A function with two modes to control transforming input values where parts of the input should
        remain unchanged. If a single input is given, the function will return (the modifiable sections, unmodifiable sections).
        If two inputs are given, it is assume they are in the order (the modifiable sections, unmodifiable sections) and the
        function will fuse them together.
        """

        super(CategoryTransformer, self).__init__(subtransformer_args, input_constraints, input_processor)

    def init_transformation_record(self, x):
        """
        Define the structure of the transformation record and initialize it.
        :param x: The input value to be transformed

        :return:
        """

        return None

    def _enforce_constraints(self, x, transformation_record):
        """
        Not Used
        :param x: Input value to be modified
        :param transformation_record:

        :return: A "clipped" input and corrected transformation record
        """

        return x, transformation_record

    def _is_possible(self, x, transformation_record=None, transformation_value=None, *args, **kwargs):
        """
        Determines if the action specified by transformation_value can be performed on the current input. If it can, it returns
        True and provides a list of the arguments to use with transform().
        :param x: Input value
        :param transformation_record: A record tracking the transformation already applied to x.
        :param transformation_value: The definition of the transformation to apply
        :return: True if possible, arguments to use for transform()
        """
        if transformation_value is None:
            return False, []  # Can't verify default actions

        possible, action_args, transformation_effect = self.subtransformer_list[transformation_value[0]].is_possible(
            x, *transformation_value[1:]
        )

        if not possible:
            return False, []

        return True, action_args



from uret.transformers import Transformer

import numpy as np
import random


class NumberTransformer(Transformer):
    name = "NumberTransformer"

    def __init__(self, subtransformer_args, input_constraints={}, input_processor=None, number_type="float"):
        """
        Initialize a `NumberTransformer` object.
        :param subtransformer_list: The functions to be used to transform the input of a certain data type.
        :param input_constraints: The constraints to enforce on the input after transformation. The possible enforcement values
        are:
            - "bounds": Should be in the form (lower, upper, method). Mix and max indicate the end points of possible values for
            the input. Method is an optional parameter, which can control the bound clipping behavior. The possible values are:
                0: Always clip the values based on the mix/max bounds. (Default)
                1: Ensure that the value never drops below the base feature.
                2: Ensure that the value never goes above the base feature.
                3: Ensure that the value never drops below the base feature and the value is not clipped up to the lower bound.
                   In other words, never clip up from the base value
                4: Ensure that the value never goes above the base feature and the value is not clipped down to the upper bound.
                   In other words, never clip down from the base value
                5: A combination of 3 and 4. In other words, never clip in from the base value
            - "eps": Should be in the form {"value", "type"}. eps_type is an optional parameter, which controls how the
                     distance value is enforced. The possible type values are
                "abs": The distance value represents the absolute maximum value of the transformation record. (Default)
                "rel": The distance value represents a percentage of the original value should be used to enforce
                       constraints.
        :param input_processor: A function with two modes to control transforming input values where parts of the input should
        remain unchanged. If a single input is given, the function will return (the modifiable sections, unmodifiable sections).
        If two inputs are given, it is assume they are in the order (the modifiable sections, unmodifiable sections) and the
        function will fuse them together.
        :param number_type: specifies the type of number that is being modified. Currently supports "int" or "float"
        """
        self.number_type = number_type.lower()
        for key in input_constraints:
            if key not in ["bounds", "eps"]:
                raise ValueError(key + " is an unrecognized constraint type")
            elif key == "bounds" and "method" not in input_constraints[key].keys():
                input_constraints[key]["method"] = 0
            elif key == "eps":
                if isinstance(input_constraints[key], int) or isinstance(
                    input_constraints[key], float
                ):  # If not provided, it is asumed to be abs
                    input_constraints[key] = {}
                    input_constraints[key]["value"] = temp_val
                    input_constraints[key]["type"] = "abs"
                elif "type" not in input_constraints[key].keys():
                    input_constraints[key]["type"] = "abs"
                elif "value" not in input_constraints[key].keys():
                    raise ValueError("Value must be provided for eps constraint")
                elif input_constraints[key]["type"] == "rel" and abs(input_constraints[key]["value"]) > 1:
                    raise ValueError("Illegal value. With relative eps, the value must be in the range (0,1]")

        input_constraints["eps"]["value"] = abs(input_constraints["eps"]["value"])

        for args in subtransformer_args:
            if "init_args" in args.keys():
                args["init_args"][
                    "number_type"
                ] = number_type  # Number type tells the subtransformers what data type they are modifying

        super(NumberTransformer, self).__init__(subtransformer_args, input_constraints, input_processor)

    def init_transformation_record(self, x):
        """
        Define the structure of the transformation record and initialize it.
        :param x: The input value to be transformed

        :return: A record tracking the transformation already applied to x. This is used for constraint
        enforcement and action validation. The definition of the transformation record depends on the Transformer's possible input
        constraints
        """

        return 0

    def _enforce_constraints(self, x, transformation_record):
        """
        Enforce predefined input constraints
        :param x: Input value to be modified
        :param transformation_record: A record tracking the transformations already applied to x. For numbers, it is a int/float
        distance from the original value

        :return: A "clipped" input and corrected trasnformation record
        """

        original_value = x - transformation_record
        clipped_x = x
        if "eps" in self.input_constraints:
            # Absolute clipping - Clip based on a flat amount from the original value
            if self.input_constraints["eps"]["type"] == "abs":
                clipped_x = np.clip(
                    clipped_x,
                    original_value - self.input_constraints["eps"]["value"],
                    original_value + self.input_constraints["eps"]["value"],
                )
            # relative clipping - Clip based on a percentage of the original value
            else:
                clipped_x = np.clip(
                    clipped_x,
                    original_value - (original_value * self.input_constraints["eps"]["value"]),
                    original_value
                    + (self.input_constraints["bounds"].get("upper") * self.input_constraints["eps"]["value"]),
                )
        if "bounds" in self.input_constraints:
            method = self.input_constraints["bounds"]["method"]

            # For the following code, we assume the base features are within normal input bounds
            # 0: Always clip the values based on the mix/max bounds. (Default)
            # 1: Ensure that the value never drops below the base feature.
            # 2: Ensure that the value never goes above the base feature.
            # 3: Ensure that the value never drops below the base feature and the value is not clipped up to the lower bound.
            #    In other words, never clip up from the base value
            # 4: Ensure that the value never goes above the base feature and the value is not clipped down to the upper bound.
            #    In other words, never clip down from the base value
            # 5: A combination of 3 and 4. In other words, never clip in from the base value
            
            # 0
            lower_bound = self.input_constraints["bounds"].get("lower")
            upper_bound = self.input_constraints["bounds"].get("upper")

            if method == 1 or method == 3 or method == 5:
                lower_bound = original_value
                if (method == 3 or method == 5) and self.input_constraints["bounds"].get("lower") is not None:
                    lower_bound = min(lower_bound, self.input_constraints["bounds"].get("lower"))
                # Ensure that the value never goes above the base feature. -2 also means that ensure you don't clip down to the bound if the base feature is higher
            elif method == 2 or method == 4 or method == 5:
                upper_bound = original_value
                if (method == 4 or method == 5) and self.input_constraints["bounds"].get("upper") is not None:
                    upper_bound = max(upper_bound, self.input_constraints["bounds"].get("upper"))

            clipped_x = np.clip(clipped_x, lower_bound, upper_bound)
        corrected_transformation_record = transformation_record + (clipped_x - x)
        return clipped_x, corrected_transformation_record

    def _is_possible(self, x, transformation_record=None, transformation_value=None):
        """
        Determines if the action specified by transformation_value can be performed on the current input. If it can, it returns
        True and provides a list of the arguments to use with transform().
        :param x: Input value
        :param transformation_record: A record tracking the transformation already applied to x. This is used for constraint
        enforcement and action validation
        :param transformation_value: The definition of the transformation to apply
        :return: True if possible, arguments to use for transform()
        """

        if transformation_value is None:
            return False, []  # Can't verify default actions

        possible, action_args, transformation_effect = self.subtransformer_list[transformation_value[0]].is_possible(
            x, *transformation_value[1:]
        )

        if not possible:
            return False, []

        # Validates the action with respect to the input constraints
        original_value = x - transformation_record
        # Action is invalid if the transformation record is at an eps boundary and the action's effect would move past it
        if "eps" in self.input_constraints:
            # Absolute clipping - Clip based on a flat amount from the original value.
            if self.input_constraints["eps"]["type"] == "abs":
                max_delta = self.input_constraints["eps"]["value"]
            # relative clipping - Clip based on a percentage of the original value.
            else:
                max_delta = (x - transformation_record) * self.input_constraints["eps"]["value"]

            if (transformation_effect > 0 and transformation_record >= max_delta) or (
                transformation_effect < 0 and transformation_record <= -1 * max_delta
            ):
                return False, []

        # Action is invalid if x is at an boundary for the range of valid values and the action's effect would move past it
        if "bounds" in self.input_constraints:
            method = self.input_constraints["bounds"]["method"]

            # For the following code, we assume the base features are within normal input bounds
            # 0: Always clip the values based on the mix/max bounds. (Default)
            # 1: Ensure that the value never drops below the base feature.
            # 2: Ensure that the value never goes above the base feature.
            # 3: Ensure that the value never drops below the base feature and the value is not clipped up to the lower bound.
            #    In other words, never clip up from the base value
            # 4: Ensure that the value never goes above the base feature and the value is not clipped down to the upper bound.
            #    In other words, never clip down from the base value
            # 5: A combination of 3 and 4. In other words, never clip in from the base value
            
            # 0
            lower_bound = self.input_constraints["bounds"].get("lower")
            upper_bound = self.input_constraints["bounds"].get("upper")
            
            if method == 1 or method == 3 or method == 5:
                lower_bound = original_value
                if (method == 3 or method == 5) and self.input_constraints["bounds"].get("lower") is not None:
                    lower_bound = min(lower_bound, self.input_constraints["bounds"].get("lower"))
                # Ensure that the value never goes above the base feature. -2 also means that ensure you don't clip down to the bound if the base feature is higher
            elif method == 2 or method == 4 or method == 5:
                upper_bound = original_value
                if (method == 4 or method == 5) and self.input_constraints["bounds"].get("upper") is not None:
                    upper_bound = max(upper_bound, self.input_constraints["bounds"].get("upper"))
                

            if (transformation_effect > 0 and x >= upper_bound) or (transformation_effect < 0 and x <= lower_bound):
                return False, []

        return True, action_args



from abc import ABC, abstractmethod

import numpy as np
import random
import tqdm


def create_default_loss_func(scoring_alg, feature_extractor, model_predict, target_label=None):
    """
    Creates a scoring function based on the loss type
    """

    if scoring_alg == "model_loss":

        def loss_function(x, original_pred):
            """
            Default loss function that returns the probability of the current input for the correct
            label or for the target label if provided.
            param original_pred: Prediction on the original input before any modification
            """
            model_prediction = model_predict(feature_extractor(x))
            if len(np.shape(model_prediction)) == 2:
                model_prediction = model_prediction[0]

            if len(model_prediction) != 1:
                small_noise = random.uniform(0.00001, 0.000001)  # Avoid divide by zero or all/nothing classifiers
                if target_label is None:
                    return np.log(model_prediction[np.argmax(original_pred)] + small_noise)
                else:
                    return -np.log(model_prediction[target_label] + small_noise)
            elif not isinstance(model_prediction, float):
                small_noise = random.uniform(0.001, 0.0001)  # Compensate for lack of prob
                if target_label is None:
                    if model_prediction[0] == original_pred[0]:
                        print("a")
                        return np.log(1 - small_noise)
                    else:
                        print("b")
                        return np.log(small_noise)
                else:
                    if model_prediction[0] == target_label:
                        return np.log(small_noise)
                    else:
                        return np.log(1 - small_noise)
            else:
                raise ValueError(
                    f"Model predict function appears to return a single float value. Please provide a custom model loss function as there is no default implementation for this case"
                )

    else:
        import scipy.spatial

        def loss_function(x, target_feature):
            """
            Default loss function that returns the similarity of the current input's features with respect
            to the target features
            param target_feature: Target features to attain
            """
            return 1-scipy.spatial.distance.cosine(feature_extractor(x), target_feature)

    return loss_function


class GraphExplorer(ABC):
    """
    Abstract base class for a graph explorer. This object explores the transformation graph and returns possibly adversarial samples
    """

    def __init__(
        self,
        model_predict,
        ranking_algorithm,
        feature_extractor=lambda x: x,
        scoring_alg="model_loss",
        scoring_function=None,
        dependencies=None,
        target_label=None,
    ):
        """
        Create a GraphExplorer

        param model_predict: Classification model that the adversarial sample is being generated for. This should just be the callable prediction function of the model. It is assume it outputs a 1-hot vector or a class probability vector
        param ranking_algorithm: Algorithm used to estimate the value of the neighboring vertices
        param feature_extractor: Algorithm that converts a raw input into a input that the model can ingest
        param scoring_alg: What type of loss will be used to score a vertex. This informs the explorer what parameters should be passed to the scoring function
        param scoring_function: Function used to score a vertex (i.e. input sample). This should take in 2 arguments and return a value in which, the lower the value, the "better" the index at achieving the goal
        param dependencies: A list of functions in the form [callable, callable_args] that enforce input dependencies (e.g. features 1 and 2 should always add up to feature 3)
        param target_label: Set this value to a class index in the model's output range to perform a targeted attack. If none, it is assume that we are performing an untargeted attack by default.
        """

        self.model_predict = model_predict
        self.ranking_algorithm = ranking_algorithm

        self.feature_extractor = feature_extractor

        self.scoring_alg = scoring_alg.lower()
        self.scoring_function = scoring_function
        self.target_label = target_label

        if not self.scoring_alg == "model_loss" and not self.scoring_alg == "feature_loss":
            raise ValueError(self.scoring_alg + "is an unsupported loss type")

        if scoring_function is None:
            if (
                self.scoring_alg == "model_loss"
                and self.target_label is not None
                and not isinstance(self.target_label, int)
            ):
                raise ValueError("The target label must be a single int if provided.")
            scoring_function = create_default_loss_func(
                self.scoring_alg, self.feature_extractor, self.model_predict, self.target_label
            )
        self.scoring_function = scoring_function

        if dependencies is None:
            dependencies = []
        if len(dependencies) != 0 and not isinstance(dependencies[0], list):
            raise ValueError(f"Dependencies is expected to be a list of lists")
        self.dependencies = dependencies

    @abstractmethod
    def search(self, sample, score_input, *args, **kwargs):
        """
        Run the search algorithm for the sample and return the transformed sample.
        sample: The sample to transform
        score_input: An input to the scoring function used to rank an edge

        return: transformed_sample, transformation_record, estimated score of transformed sample
        """
        raise NotImplementedError

    def explore(self, x, target_features=None, return_record=False):
        """
        Generate adversarial sample(s) for x using the algorithms created during init.
        param target_features: If using feature loss, then this value contains the target features.
        param return_record: A boolean that if True will return the transformation record as well.
        """

        if self.scoring_alg == "feature_loss":
            if target_features is None:
                raise ValueError("The target features must be provided in order to use feature_loss")
            elif len(target_features) != len(x):
                raise ValueError("There must be as many target features as inputs")

        generated_samples = []

        if return_record:
            records = []

        for i, sample in enumerate(tqdm.tqdm(x)):
            original_pred = self.model_predict(self.feature_extractor(sample))
            if len(np.shape(original_pred)) == 2:
                original_pred = original_pred[0]

            best_sample = None
            best_score = np.inf
            if return_record:
                best_record = None

            if self.scoring_alg == "model_loss":
                score_input = original_pred
            else:
                score_input = target_features[i]

            for sample_next, transformation_record, _ in self.search(sample, score_input):

                # Score the current sample
                score = self.scoring_function(sample_next, score_input)

                # Early exit conditions
                # If using feature loss, then we can early exit once the target features are attained
                # Maybe consider a "close-enough" condition instead?
                if self.scoring_alg == "feature_loss" and np.array_equal(
                    target_features[i], self.feature_extractor(sample_next)
                ):
                    best_sample = sample_next
                    best_score = score
                    break

                # For all loss types, we can early exit if an adversarial example is found
                new_prediction = self.model_predict(self.feature_extractor(sample_next))
                if len(np.shape(new_prediction)) == 2:
                    new_prediction = new_prediction

                if self.target_label is not None and np.argmax(new_prediction) == self.target_label:
                    best_sample = sample_next
                    best_score = score
                    if return_record:
                        best_record = transformation_record
                    break
                elif np.argmax(new_prediction) != np.argmax(original_pred):
                    best_sample = sample_next
                    best_score = score
                    if return_record:
                        best_record = transformation_record
                    break

                # Check if the current sample is better
                if best_sample is None or score < best_score:
                    best_sample = sample_next
                    best_score = score
                    if return_record:
                        best_record = transformation_record

            if return_record:
                records.append(best_record)

            generated_samples.append(best_sample)

        if return_record:
            return generated_samples, records

        return generated_samples

    def _enforce_dependencies(self, sample):
        """
        Enforces dependencies specified in the list. Each element in dependencies should be of the form
        [callable_dependency function, dependency arguments]
        """
        for d in self.dependencies:
            sample = d[0](sample, **d[1])
        return sample

    def train(self, training_samples, value_function=None, value_input_function=None, **kwargs):
        """
        Call the training algorithm of the ranking algorithm
        """

        ## Create value functions if undefined
        if value_function is None and value_input_function is None:
            if self.scoring_alg == "model_loss":

                def value_function(x, original_pred):
                    """
                    Default value function that returns the change in probability for a given label.
                    :param original_pred: Prediction on the original input before any modification
                    """

                    model_prediction = self.model_predict(self.feature_extractor(x))
                    if len(np.shape(model_prediction)) == 2:
                        model_prediction = model_prediction[0]

                    if self.target_label is None:
                        target_label = np.argmax(original_pred)
                        return model_prediction[target_label] - original_pred[target_label]
                    # We swap the order for target label as we want lower values to reflect "good" changes when ranking edges based on how scores are processed. In the targeted label case, a "good" change is when the target label's prob increases
                    else:
                        target_label = self.target_label
                        return original_pred[target_label] - model_prediction[target_label]

            else:

                def value_function(x, original_feature):
                    """
                     Default value function that returns the change in feature for a given label
                    :param original_pred: Prediction on the original input before any modification
                    """
                    return self.feature_extractor(x) - original_feature

            def value_input_function(x):
                """
                Computes the value representation for an input
                """
                if self.scoring_alg == "model_loss":
                    model_prediction = self.model_predict(self.feature_extractor(x))
                    if len(np.shape(model_prediction)) == 2:
                        model_prediction = model_prediction[0]
                    return model_prediction
                else:
                    return self.feature_extractor(x)

        elif value_function is None or value_input_function is None:
            raise ValueError("Value_function and value_input_function must both be defined")

        self.ranking_algorithm._train(
            training_samples,
            value_function,
            value_input_function,
            self.scoring_alg,
            dependencies=self.dependencies,
            **kwargs,
        )



from uret.transformers import SubTransformer

import random
import re
import string

from copy import deepcopy


class Substitution(SubTransformer):
    name = "Substitution"

    def __init__(self, transformation_range=None, max_actions=0, subtransformer_index=None):
        """
        Initialize a `Substitution` object. This object substitutes a character in a string.
        :param transformation_range: The characters that can be deleted.
        :param max_actions: If > 0, this determines how many actions are returned when get_action_list is called
        :param subtransformer_index: The index of the subtransformer in the transformer list. If not none, the subtransformer assumes the transformation record 'actions_taken' value is a list.
        """
        if transformation_range is None:
            self.transformation_range = string.ascii_lowercase + string.digits
        else:
            self.transformation_range = transformation_range

        self.max_actions = max_actions
        self.subtransformer_index = subtransformer_index

        super(Substitution, self).__init__()

    def transform(self, x, transformation_record, transformation_value, index=None):
        """
        Perform a substitution on input x based on the remaining input arguments
        :param x: Input text
        :param transformation_value: (Target value, new value) pair
        :param index: Index of the character to make the substitution. If not provied, a random matching position will be used
        :return: A transformed input and the modified record
        """

        if transformation_value is not None:
            target_value, new_value = transformation_value
        else:
            target_value = random.choice([v for v in x if v in self.transformation_range])
            new_value = None

        if target_value not in self.transformation_range or (
            new_value is not None and new_value not in self.transformation_range
        ):
            raise ValueError("Transformation value is not in the pre-defined range")

        if new_value is None:
            new_value = random.choice(self.transformation_range.replace(target_value, ""))

        if index is None:
            replace_inds = [match.start() for match in re.finditer(target_value, x)]

            if len(replace_inds) == 0:
                print(target_value, "not present")
                return x
            replacement_pos = random.choice(replace_inds)

        elif x[index : index + len(new_value)] == target_value:
            replacement_pos = index

        else:
            raise ValueError(transformation_value + " wasn't found at the specified index")

        new_x = x[:replacement_pos] + new_value + x[replacement_pos + len(new_value) :]

        transformation_record["prev_state"] = deepcopy(transformation_record["current_state"])
        if self.subtransformer_index is not None:
            transformation_record["current_state"]["actions_taken"][self.subtransformer_index] += 1
        else:
            transformation_record["current_state"]["actions_taken"] += 1

        transformation_record["current_state"]["value"] = new_x

        return new_x, transformation_record

    def get_action_list(self):
        """
        Return a list of characters pairs that the subtransformer can replace
        """
        action_list = [
            (old, new) for new in self.transformation_range for old in self.transformation_range if old != new
        ]

        if self.max_actions > 0:
            action_list.shuffle()
            return action_list[: self.max_actions]
        return action_list

    def is_possible(self, x, transformation_value):
        """
        Determines if the action specified by tranformation_value can be performed on the current input. If it can, it returns
        True and provides a list of the arguments to use with transform().
        :param x: Input value
        :param transformation_value: The definition of the transformation to apply

        :return: True if possible, arguments to use for transform(), and the change in string length
        """
        target_value, new_value = transformation_value
        if target_value not in self.transformation_range or (
            new_value is not None and new_value not in self.transformation_range
        ):
            return False, None, 0

        replace_inds = [match.start() for match in re.finditer(target_value, x)]

        if len(replace_inds) == 0:
            return False, None, 0

        return True, [[transformation_value, ind] for ind in replace_inds], 0

    def get_possible(self, x):
        """
        Determines what actions can be performed on the input. For string, only characters in the string can be replaced
        :param x: Input Value

        :return: A list of transformation values
        """
        uniq_chars = [c for c in list(set(x)) if c in self.transformation_range]
        possible_actions = [(old, new) for new in self.transformation_range for old in uniq_chars if old != new]

        return possible_actions



from uret.transformers import SubTransformer
from uret.transformers.binary.subtransformers import binary_to_bytez

import random
import numpy as np

from copy import deepcopy


class OverlayAppend(SubTransformer):
    name = "OverlayAppend"

    def __init__(
        self, length=[6, 16], distribution_values=[0, 126, 255], max_actions=10, seed=None, subtransformer_index=None
    ):
        """
        Initialize a `OverlayAppend` object. This object adds bytes to the binary
        :param length: either an int representing the maximum append length or a pair of its represetning the (lower, upper) append
        length.
        :param distribution_values: A list of ints, which controls the possible byte values to be appended to the binary.
        :param max_actions: This determines how many actions are returned when actions are generated
        :param seed: Random seed value
        :param subtransformer_index: An int indicating where in the transformation record the subtransformer is defined. This is used to enforce "max # of transformations for this subtransformer" constraint.
        """
        if isinstance(length, list):
            self.min_val = length[0]
            self.max_val = length[1]
        else:
            self.min_val = 1
            self.max_val = length[1]

        self.distribution_values = distribution_values

        if max_actions <= 0:
            raise ValueError("max_actions must be greater than 0")

        self.max_actions = max_actions
        self.seed = seed
        self.subtransformer_index = subtransformer_index

        super(OverlayAppend, self).__init__()

    def transform(self, x, transformation_record, transformation_value, length=None):
        """
        Add bytes to the binary
        :param x: Input Value
        :param transformation_record: Record of changes
        :param transformation_value: Distribution to draw from when creating context
        :param length: length of content to append
        :return: A transformed input and modified transformation record
        """
        import lief # lgtm [py/repeated-import]

        random.seed(self.seed)

        if transformation_value is None:
            transformation_value = random.choice(self.distribution_values)
        if length is None:
            length = random.randint(self.min_val, self.max_val)

        # choose the upper bound for a uniform distribution in [0,upper]
        # upper chooses the upper bound on uniform distribution:
        # distribution_value =0 would append with all 0s
        # distribution_value =126 would append with "printable ascii"
        # distribution_value =255 would append with any character
        new_x = x + bytes([random.randint(0, transformation_value) for _ in range(length)])

        # update the transformation record
        transformation_record["prev_state"] = deepcopy(transformation_record["current_state"])
        if self.subtransformer_index is not None:
            transformation_record["current_state"]["actions_taken"][self.subtransformer_index] += 1
        else:
            transformation_record["current_state"]["actions_taken"] += 1

        transformation_record["current_state"]["value"] = new_x

        return new_x, transformation_record

    def get_action_list(self):
        """
        Return a list of (length, distribution_value) pairs. This will always return a list at least as long as the
        "distribution_value" list even if max actions is smaller.
        """

        action_list = list(self.distribution_values)

        return action_list

    def is_possible(self, x, transformation_value):
        """
        Determines if the action specified by transformation_value can be performed on the current input.
        :param x: Input value
        :param transformation_value: The definition of the transformation to apply

        :return: If action is possible and the transformation value. The effect on the record is not used
        """
        lengths = list(range(self.min_val, self.max_val + 1))

        if self.max_actions > 0:
            random.shuffle(lengths)
            lengths = lengths[: self.max_actions]

        return True, [[transformation_value, l] for l in lengths], None

    def get_possible(self, x):
        """
        Determines what actions can be performed on the input.
        :param x: input_value

        :return: A list of transformation values
        """

        return self.get_action_list()



import numpy as np
import random

import rl
import rl.core
from rl.memory import SequentialMemory
from rl.random import OrnsteinUhlenbeckProcess
from rl.agents.dqn import DQNAgent
from rl.policy import Policy, LinearAnnealedPolicy, EpsGreedyQPolicy
from rl.callbacks import ModelIntervalCheckpoint

import tensorflow as tf
import keras


class RLEnv(rl.core.Env):
    def __init__(
        self,
        task,
        input_encoding,
        # fe,
        # state_x_dimension,
        # state_f_dimension=20,
        max_steps=np.inf,
    ):
        """
        @input fe: feature extractor
        @input x_train: raw training data used as input to fe
        """
        self.flattened_transformers = task.flattened_transformers

        self.task = task
        self.input_encoding = input_encoding

        # self.fe = fe
        # self.x_initial = None
        # self.x_target = None
        self.x_current = None
        # self.x_train = None
        # self.f_target = None
        self.num_steps = 0
        self.reward_prev = 0

        # self.norm_p = norm_p
        # self.state_x_dimension = state_x_dimension
        # self.state_f_dimension = state_f_dimension
        # self.observation_dimension = self.state_x_dimension + self.state_f_dimension
        self.max_steps = max_steps
        self.max_actions_applied = task.max_actions_applied
        self.num_action_kinds = task.num_action_kinds
        self.verbose = 0

        self.y_num_actions_applied = 0
        self.y_action_sequence = [0]

        self.task.__iter__()

    # def set_x_train(self, x_train):
    #     self.x_train = x_train
    #     self.x_initial = None
    #     self.x_target = None
    #     self.x_current = None
    #     self.f_target = None
    #     self.num_steps = 0
    #     self.reward_prev = 0

    def set_verbosity(self, verbose=0):
        self.verbose = verbose

    def step(self, action):
        x_prev = self.x_current
        self.x_current = self.action(action, self.x_current)
        r0 = self.reward_prev
        self.reward_prev = r1 = self.reward(self.x_current)
        reward = r1 - r0

        state = self.get_state(self.x_current)
        stop = self.is_complete(r1)
        self.num_steps += 1
        if self.verbose > 1:
            print(
                "\nStep: %s ==> %s (action=%d, reward=%f, %s)"
                % (x_prev, self.x_current, action, reward, "success" if self.x_current == self.x_target else "not yet")
            )
        if self.num_steps > self.max_steps:
            stop = True
        return state, reward, stop, {}

    def get_optimal_action(self):
        if self.num_steps < len(self.y_action_sequence):
            if self.verbose > 1:
                print(
                    "Env::get_optimal_action(policy)::num_steps=",
                    self.num_steps,
                    ", y_action_sequence=",
                    self.y_action_sequence,
                )
            return self.y_action_sequence[self.num_steps]
        else:
            if self.verbose > 1:
                print(
                    "Env::get_optimal_action(rand)::num_steps=",
                    self.num_steps,
                    ", y_action_sequence=",
                    self.y_action_sequence,
                )
            return np.random.randint(0, self.num_action_kinds)

    def get_state(self, xi):
        state = np.concatenate([self.input_encoding(xi), self.task.get_delta_target(xi)], axis=0)
        return state

    def reset(self):
        ## TODO: Training time task has different goal to the test time task.
        # In training, the target and the different are measured against x_target.
        # Can we change Task to accomodate external goals? Or is it ideal to make a RLTask?
        # self.x_target = self.x_initial = next(self.task)  # Maybe, we just need to take x_init, x_target from the task, and use these instead of Task itself. But then, can we use the score function?
        # self.y_num_actions_applied = random.randrange(1, self.max_actions_applied)
        # a_seq = []
        # for i in range(self.y_num_actions_applied):
        #     a = random.randrange(self.num_action_kinds)
        #     self.x_target = self.action(a, self.x_target)
        #     a_seq.append(a)
        # self.y_action_sequence = a_seq
        # if self.verbose > 1:
        #     print("\n  " + self.x_initial + "=(%d)=>" % self.y_num_actions_applied + self.x_target)

        self.x_current, self.y_action_sequence, self.y_num_actions_applied = next(self.task)
        # self.f_target = self.fe(self.x_target)
        self.num_steps = 0

        # Initial reward
        # f_current = self.fe(self.x_current)
        # state = np.concatenate([self.input_encoding(self.x_current), self.f_target-f_current], axis=0)
        # state = self.get_state(self.x_current, self.f_target)
        state = self.get_state(self.x_current)
        self.reward_prev = self.reward(self.x_current)
        return state

    def render(self, mode="human"):
        print("\n=>" + self.x_current + "(%f)" % self.reward_prev)

    def close(self):
        pass

    def action(self, action, x):
        """
        Applies an action specified by action on x and returns it.
        :return:
        :rtype:
        """
        t, a = self.flattened_transformers[action]
        x = t.transform(x, a)
        return x

    # def input_encoding(self, x):  # Need to be customized. Model-specific mapping to a vector.
    #     """
    #     Encodes the raw input x into a numerical form.
    #     :return:
    #     :rtype:
    #     """
    #     raise NotImplementedError

    def reward(self, x):
        """
        Returns the reward of x for the current task.
        :return:
        :rtype:
        """
        return self.task.score(x)

    def is_complete(self, r):  # Need to be customized
        """
        Returns if the reward r is its maximum score and needs to end the episode.
        :return:
        :rtype:
        """
        return r == 1


class DGAEnv(RLEnv):
    character_to_int = {}
    character_to_int.update(zip("abcdefghijklmnopqrstuvwxyz", range(26)))
    character_to_int.update(zip("0123456789", range(26, 36)))
    character_to_int["-"] = 36
    character_to_int["."] = 37

    int_to_character = {v: k for (k, v) in character_to_int.items()}

    num_inputs = 64  # 64 characters
    num_action_kinds_list = (
        1,
        12,
        len(character_to_int),
    )  # 1 action (substitution), 12 locations, 37 characters (alphabet, numbers, -)
    num_hidden = 32

    num_action_kinds = num_action_kinds_list[0] * num_action_kinds_list[1] * num_action_kinds_list[2]

    def action(self, action_encoded, x):  # Need to be customized.
        action = action_encoded // (DGAEnv.num_action_kinds_list[1] * DGAEnv.num_action_kinds_list[2])
        action_location = (
            action_encoded
            % (DGAEnv.num_action_kinds_list[1] * DGAEnv.num_action_kinds_list[2])
            // DGAEnv.num_action_kinds_list[2]
        )
        action_character = (
            action_encoded
            % (DGAEnv.num_action_kinds_list[1] * DGAEnv.num_action_kinds_list[2])
            % DGAEnv.num_action_kinds_list[2]
        )

        if action == 0:  # substitution
            dn_new = x[:action_location] + DGAEnv.int_to_character[action_character] + x[action_location + 1 :]
        return dn_new

    def input_encoding(self, x):  # Need to be customized
        state = np.zeros([self.state_x_dimension])
        state[: len(x)] = list(map(lambda c: DGAEnv.character_to_int[c], x))
        return state

    def reward(self, x, f_target):  # Need to be customized
        f = self.fe(x)
        dist = np.linalg.norm(f - f_target, self.norm_p)  # Now we have to choose the norm.
        return len(f_target) - dist  # Linear reward. len(target) is an arbitrary max value.

    def is_complete(self, r):  # Need to be customized
        return r == len(self.f_target)


# Policies


class GoldStandardPolicy(Policy):
    def __init__(self, env, is_training, on_policy):
        #         self.action_random_select = True
        self.is_training = is_training
        self.on_policy = on_policy
        self.env = env

    def select_action(self, q_values):
        if self.is_training == True and self.on_policy == True:
            action_triple = self.env.get_optimal_action()
            return action_triple
        else:
            return np.argmax(q_values)


class GSEpsGreedyQPolicy(GoldStandardPolicy):
    def __init__(self, env, is_training, on_policy, eps=0.1, gs=0.4):
        super(GSEpsGreedyQPolicy, self).__init__(env, is_training, on_policy)
        self.eps = eps
        self.gs = gs

    def select_action(self, q_values):
        """Return the selected action
        # Arguments
            q_values (np.ndarray): List of the estimations of Q for each action
        # Returns
            Selection action
        """
        assert q_values.ndim == 1
        nb_actions = q_values.shape[0]

        p = np.random.uniform()
        if p < self.eps:
            action = np.random.randint(0, nb_actions)
        elif p >= self.eps and p < self.eps + self.gs:
            action = super(GSEpsGreedyQPolicy, self).select_action(q_values)
        else:
            action = np.argmax(q_values)
        return action

    def get_config(self):
        config = super(GSEpsGreedyQPolicy, self).get_config()
        config["eps"] = self.eps
        config["gs"] = self.gs
        return config


def train(
    model,
    env,
    nb_steps=1750000,
    log_interval=10000,
    nb_max_episode_steps=3,
    checkpoint_name=None,
    checkpoint_interval=10000,
    linear_anneal_nb_steps=1000000,
):
    # Train
    policy = LinearAnnealedPolicy(
        GSEpsGreedyQPolicy(env, True, True, eps=0.1, gs=0.3),
        attr="eps",
        value_max=1.0,
        value_min=0.1,
        value_test=0.05,
        nb_steps=linear_anneal_nb_steps,
    )
    memory = SequentialMemory(limit=100000, window_length=1)
    if hasattr(model.output, "__len__"):  # Keras-RL & TF incompatibility workaround
        delattr(tf.Tensor, "__len__")

    dqn = DQNAgent(
        model=model,
        nb_actions=env.num_action_kinds,
        policy=policy,
        memory=memory,
        processor=None,
        nb_steps_warmup=100,
        gamma=0.99,
        target_model_update=10000,
        train_interval=1,
        delta_clip=1.0,
    )
    # dqn.compile(Adam(lr=.00025), metrics=['mae'])
    dqn.compile(keras.optimizers.Adam(), metrics=["mae"])

    callbacks = []
    if checkpoint_name:
        callbacks.append(ModelIntervalCheckpoint(checkpoint_name, checkpoint_interval))

    env.set_verbosity(0)
    dqn.fit(
        env,
        nb_steps=nb_steps,
        log_interval=log_interval,
        nb_max_episode_steps=nb_max_episode_steps,
        callbacks=callbacks,
    )
    return dqn




import array

def binary_to_bytez(binary, dos_stub=False, imports=False):
    import lief # lgtm [py/repeated-import]

    # Write modified binary to disk
    builder = lief.PE.Builder(binary)
    builder.build_imports(imports)
    builder.build()

    bytez = array.array("B", builder.get_build()).tobytes()
    return bytez



from uret.transformers import SubTransformer
from uret.transformers.binary.subtransformers import binary_to_bytez

import random
from copy import deepcopy


class SectionRename(SubTransformer):
    name = "SectionRename"

    def __init__(self, section_names=None, seed=None, subtransformer_index=None):
        """
        Initialize a `SectionRename` object. This object renames a section.
        :param section_names: A list of strings containing section names
        :param seed: Random seed value
        :param subtransformer_index: An int indicating where in the transformation record the subtransformer is defined. This is used to enforce "max # of transformations for this subtransformer" constraint.
        """
        self.seed = seed

        if section_names is None:
            self.section_names = (
                open("agrex/transformers/binary/subtransformers/section_names.txt", "r").read().rstrip().split("\n")
            )

        self.seed = seed
        self.subtransformer_index = subtransformer_index

        super(SectionRename, self).__init__()

    def transform(self, x, transformation_record, transformation_value=None):
        """
        Add bytes to the binary
        :param x: Input Value
        :param transformation_record: Record of changes
        :param transformation_value: (length, distribution) to append
        :return: A transformed input and modified transformation record
        """
        import lief # lgtm [py/repeated-import]

        random.seed(self.seed)
        binary = lief.PE.parse(list(x))

        if transformation_value is None:
            transformation_value = random.choice(self.section_names)

        # Pick a random section to rename
        target_section = random.choice(binary.sections)
        if target_section.name != transformation_value:
            target_section.name = transformation_value
            new_x = binary_to_bytez(binary)

            # update the transformation record
            transformation_record["prev_state"] = deepcopy(transformation_record["current_state"])
            if self.subtransformer_index is not None:
                transformation_record["current_state"]["actions_taken"][self.subtransformer_index] += 1
            else:
                transformation_record["current_state"]["actions_taken"] += 1

            transformation_record["current_state"]["value"] = new_x
            return new_x, transformation_record
        else:
            return x, transformation_record

    def get_action_list(self):
        """
        Return a list of section names
        """
        return self.section_names

    def is_possible(self, x, transformation_value):
        """
        Determines if the action specified by transformation_value can be performed on the current input.
        :param x: Input value
        :param transformation_value: The definition of the transformation to apply

        :return: If action is possible and the transformation value. The effect on the record is not used
        """
        return True, [transformation_value], None

    def get_possible(self, x):
        """
        Determines what actions can be performed on the input.
        :param x: input_value

        :return: A list of transformation values
        """
        return self.section_names



from uret.core.rankers.brute_force import BruteForce
from uret.core.rankers.lookup_table import LookupTable
from uret.core.rankers.external_model import ExternalModel
from uret.core.rankers.random import Random



from uret.transformers import SubTransformer

import random
import os
import tempfile
import subprocess

from copy import deepcopy


class UPXUnpack(SubTransformer):

    name = "UPXUnpack"

    def __init__(self, seed=None, subtransformer_index=None):
        """
        Initialize a `UPXUnpack` object. This object uses UPX to unpact the binary
        :param seed: The random seed used for transform
        :param subtransformer_index: An int indicating where in the transformation record the subtransformer is defined. This is used to enforce "max # of transformations for this subtransformer" constraint.
        """
        self.seed = seed
        self.subtransformer_index = subtransformer_index
        super(UPXUnpack, self).__init__()

    def transform(self, x, transformation_record, transformation_value=0):
        """
        Unpack x using UPX on input x
        :param x: Input Value
        :param transformation_record: Record of changes
        :param transformation_value: Not used
        :return: A transformed input and modified transformation record
        """
        random.seed(self.seed)
        tempfilename = os.path.join(tempfile._get_default_tempdir(), next(tempfile._get_candidate_names()))

        # dump to a temporary file
        with open(tempfilename, "wb") as outfile:
            outfile.write(x)

        with open(os.devnull, "w") as DEVNULL:
            retcode = subprocess.call(
                ["upx", tempfilename, "-d", "-o", tempfilename + "_unpacked"], stdout=DEVNULL, stderr=DEVNULL
            )

        os.unlink(tempfilename)
        # Op succcessful
        if retcode == 0:
            with open(tempfilename + "_unpacked", "rb") as result:
                x = result.read()
            os.unlink(tempfilename + "_unpacked")

            # update the transformation record
            transformation_record["prev_state"] = deepcopy(transformation_record["current_state"])
            if self.subtransformer_index is not None:
                transformation_record["current_state"]["actions_taken"][self.subtransformer_index] += 1
            else:
                transformation_record["current_state"]["actions_taken"] += 1

            transformation_record["current_state"]["value"] = x

        return x

    def get_action_list(self):
        """
        Return a default argument
        """
        return [0]  # Not Used

    def is_possible(self, x, transformation_value):
        """
        Determines if the action specified by transformation_value can be performed on the current input.
        :param x: Input value
        :param transformation_value: The definition of the transformation to apply

        :return: If action is possible and the transformation value. The effect on the record is not used
        """
        # not sure how to check without actually packing
        try:
            self.transform(x, None, transformation_value)
            return True, [transformation_value], None
        except FileNotFoundError:
            return False, [], None  # not sure how to check without actually packing

    def get_possible(self, x):
        """
        Determines what actions can be performed on the input.
        :param x: input_value

        :return: A list of transformation values
        """
        # not sure how to check without actually packing
        try:
            self.transform(x, None, 0)
            return [0]
        except FileNotFoundError:
            return []  # not sure how to check without actually packing



from uret.transformers import SubTransformer
from uret.transformers.binary.subtransformers import binary_to_bytez

from copy import deepcopy


class RemoveDebug(SubTransformer):
    name = "RemoveDebug"

    directory_mapper = {
        "has_signature": lief.PE.DATA_DIRECTORY.CERTIFICATE_TABLE,
        "has_debug": lief.PE.DATA_DIRECTORY.DEBUG,
    }

    def __init__(self, sub=None):
        """
        Initialize a `RemoveDebug` object. This object removes the debug data directory .
        :param sub: An int indicating where in the transformation record the subtransformer is defined. This is used to enforce "max # of transformations for this subtransformer" constraint.
        """
        self.sub = sub

        super(RemoveDebug, self).__init__()

    def transform(self, x, transformation_record, transformation_value=0):
        """
        Removed the specified data directory
        :param x: Input Value
        :param transformation_record: Record of changes
        :param transformation_value: Not used
        :return: A transformed input and modified transformation record
        """
        import lief # lgtm [py/repeated-import]

        binary = lief.PE.parse(list(x))

        for i, e in enumerate(binary.data_directories):
            if e.type == lief.PE.DATA_DIRECTORY.DEBUG:
                break

        if e.type == lief.PE.DATA_DIRECTORY.DEBUG:
            e.rva = 0
            e.size = 0

            new_x = binary_to_bytez(binary)

            # update the transformation record
            transformation_record["prev_state"] = deepcopy(transformation_record["current_state"])
            if self.sub is not None:
                transformation_record["current_state"]["actions_taken"][self.sub] += 1
            else:
                transformation_record["current_state"]["actions_taken"] += 1

            transformation_record["current_state"]["value"] = new_x

            return new_x, transformation_record

        else:
            return x, transformation_record

    def get_action_list(self):
        """
        Returns a default argument
        """
        return [0]

    def is_possible(self, x, transformation_value):
        """
        Determines if the action specified by transformation_value can be performed on the current input.
        :param x: Input value
        :param transformation_value: The definition of the transformation to apply

        :return: If action is possible and the transformation value. The effect on the record is not used
        """
        import lief # lgtm [py/repeated-import]

        binary = lief.PE.parse(list(x))

        for i, e in enumerate(binary.data_directories):
            if e.type == lief.PE.DATA_DIRECTORY.DEBUG:
                return True, [transformation_value], None

        return False, [], None

    def get_possible(self, x):
        """
        Determines what actions can be performed on the input.
        :param x: input_value

        :return: A list of transformation values
        """
        import lief # lgtm [py/repeated-import]

        binary = lief.PE.parse(list(x))

        for i, e in enumerate(binary.data_directories):
            if e.type == lief.PE.DATA_DIRECTORY.DEBUG:
                return [0]

        return []



from uret.core.rankers.ranking_algorithm import RankingAlgorithm
from uret.core.rankers.brute_force import BruteForce

import warnings
import tqdm
import scipy.spatial
import numpy


class LookupTable(RankingAlgorithm):
    """
    This implementation uses a pre-computed lookup table. It uses the brute_force method on the training samples to built a lookup table
    """

    def __init__(self, transformer_list, multi_feature_input=False, is_trained=False, feature_distance_type="sim"):
        """
        :param transformer_list: List of available transformations to be used by the algorithm.
        :param multi_feature_input: A boolean flag indicating if each index in the transformer
            list defines a transformer(s) for a single data type in the input. If False, then it
            is assumed that the transformer list defines multiple transformers for a single input
        :param is_trained: Indicates when _train has been called at least once
        :param feature_distance_type: Function to use when determining which action to use in the lookup table if table values are
            feature perturbation vectors. Option are
                sim - Cosine similarity between perturbation vectors
                l2 - l2 distance
        """
        super().__init__(transformer_list, multi_feature_input, is_trained=False)

        self.brute_force = BruteForce(transformer_list, multi_feature_input=multi_feature_input)
        self.lookup_table = {}

        if feature_distance_type != "sim" and feature_distance_type != "l2":
            raise ValueError(f"Distance type not supported yet")

        self.feature_distance_type = feature_distance_type

    def _train(
        self,
        training_samples,
        value_function,
        value_input_function,
        scoring_alg,
        dependencies=[],
        update=False,
    ):
        """
        Create a lookup table using the training samples. We run the brute force algorithm to get scores for the training data with respect to the value_function.
        :param training_samples: Input samples used to train the action dictionary
        :param value_function: Function used to compute values to compute the value a transformation has on the input with respect to a certain loss.
        :param value_input_function: Function used to compute the initial state of the sample before transform. This is fed into the value function as an input.
        :param scoring_alg: Indicates the scoring alg used for ranking if necessary
        :param dependencies: Input dependencies that need to be enforced
        """

        if not update:
            self.lookup_table = {}

        self.scoring_alg = scoring_alg  # This informs the lookup table how to compute during edge ranking
        self.value_input_function = value_input_function  # This enables re-use during edge ranking

        # Use all of the training samples and run the brute force algorithm
        print(f"Creating Dictionary")
        for i, sample in tqdm.tqdm(enumerate(training_samples)):
            initial_state = value_input_function(sample)
            edge_transform_estimates = self.brute_force.rank_edges(
                sample,
                value_function,
                initial_state,
                dependencies=dependencies,
                current_transformation_records=None,
            )

            for indices, transformer, action, _, _, score in edge_transform_estimates:
                action_key = str(action)  # Need to convert from list to hashable type
                if self.multi_feature_input:
                    if indices[0] not in self.lookup_table.keys():
                        self.lookup_table[indices[0]] = {"input_index": indices[1]}
                    if transformer.name not in self.lookup_table[indices[0]].keys():
                        self.lookup_table[indices[0]][transformer.name] = {}
                    if action_key not in self.lookup_table[indices[0]][transformer.name].keys():
                        self.lookup_table[indices[0]][transformer.name][action_key] = {"avg_change": 0, "count": 0}

                    self.lookup_table[indices[0]][transformer.name][action_key]["count"] += 1
                    self.lookup_table[indices[0]][transformer.name][action_key]["avg_change"] += (
                        score - self.lookup_table[indices[0]][transformer.name][action_key]["avg_change"]
                    ) / self.lookup_table[indices[0]][transformer.name][action_key]["count"]

                else:
                    if transformer.name not in self.lookup_table.keys():
                        self.lookup_table[transformer.name] = {}
                    if action_key not in self.lookup_table[transformer.name].keys():
                        self.lookup_table[transformer.name][action_key] = {"avg_change": 0, "count": 0}

                    self.lookup_table[transformer.name][action_key]["count"] += 1
                    self.lookup_table[transformer.name][action_key]["avg_change"] += (
                        score - self.lookup_table[transformer.name][action_key]["avg_change"]
                    ) / self.lookup_table[transformer.name][action_key]["count"]

        print(f"Dictionary Created")
        self.is_trained = True

    def rank_edges(self, sample, scoring_function, score_input, dependencies=[], current_transformation_records=None):

        if not self.is_trained:
            raise RuntimeError(
                f"Lookup table has not been trained. Please call `GraphExplorer.train(training_samples)` first"
            )

        # Create tranformation record
        if self.multi_feature_input and current_transformation_records is None:
            current_transformation_records = [None for _ in range(len(self.transformer_list))]

        return_values = []  # This will contain the (sample_index, transformer, action args, transformed sample,
        # transformation_record of the transformed sample, score)

        for transformer_index, (transformer, input_index) in enumerate(self.transformer_list):
            # Get the possible actions
            if self.multi_feature_input:
                possible_actions = transformer.get_possible(
                    sample[input_index], transformation_record=current_transformation_records[transformer_index]
                )
                action_dict = self.lookup_table[transformer_index][transformer.name]
            else:
                possible_actions = transformer.get_possible(
                    sample, transformation_record=current_transformation_records
                )
                action_dict = self.lookup_table[transformer.name]

            for action in possible_actions:
                action_key = str(action)
                if action_key not in action_dict.keys():
                    continue

                predicted_perturbation = action_dict[action_key]["avg_change"]
                if self.scoring_alg == "model_loss":
                    estimated_score = predicted_perturbation  # In the case of probabilities, we can just look for the largest positive value
                # If it's feature loss we compute the specified metric
                elif self.feature_distance_type == "sim":
                    target_perturbation = score_input - self.value_input_function(sample)
                    estimated_score = scipy.spatial.distance.cosine(predicted_perturbation, target_perturbation)
                else:
                    estimated_score = numpy.lingalg.norm(
                        self.value_input_function(sample) + predicted_perturbation, score_input
                    )

                if self.multi_feature_input:
                    return_values.append(
                        [
                            [transformer_index, input_index],
                            transformer,
                            action,
                            None,
                            None,
                            estimated_score,
                        ]
                    )
                else:
                    return_values.append([None, transformer, action, None, None, estimated_score])

        return return_values



from uret.transformers.binary.subtransformers.binary_to_bytez import binary_to_bytez

from uret.transformers.binary.subtransformers.imports_append import ImportsAppend
from uret.transformers.binary.subtransformers.overlay_append import OverlayAppend
from uret.transformers.binary.subtransformers.remove_debug import RemoveDebug
from uret.transformers.binary.subtransformers.section_add import SectionAdd
from uret.transformers.binary.subtransformers.section_rename import SectionRename
from uret.transformers.binary.subtransformers.upx_pack import UPXPack
from uret.transformers.binary.subtransformers.upx_unpack import UPXUnpack



from uret.transformers import SubTransformer
from uret.transformers.binary.subtransformers import binary_to_bytez

import random
import os

import json
from copy import deepcopy


class ImportsAppend(SubTransformer):
    name = "ImportsAppend"

    def __init__(self, transformation_range=None, max_actions=50, seed=None, subtransformer_index=None):
        """
        Initialize a `ImportsAppend` object. This object adds a new function (and library) to the import list
        :param transformation_range: A dictionary of (libary_name, function_list) pairs. If not provided, a dictionary of common
        library imports will be used
        :param max_actions: If > 0, this determines how many actions are returned when get_action_list is called
        :param seed: Random seed value
        :param subtransformer_index: An int indicating where in the transformation record the subtransformer is defined. This is used to enforce "max # of transformations for this subtransformer" constraint.
        """

        if transformation_range is None:
            self.transformation_range = json.load(
                open("agrex/transformers/binary/subtransformers/small_dll_imports.json", "r")
            )
        elif not isinstance(transformation_range, dict):
            raise ValueError("transformation_range must be a dictionary")
        else:
            self.transformation_range = transformation_range

        if max_actions <= 0:
            raise ValueError("max_actions must be greater than 0")

        self.max_actions = max_actions
        self.seed = seed
        self.subtransformer_index = subtransformer_index

        super(ImportsAppend, self).__init__()

    def transform(self, x, transformation_record, transformation_value):
        """
        Add a new function from the library specified by the transformation value to the import list
        :param x: Input Value. bytes
        :param transformation_record: Record of changes
        :param transformation_value: libary to append
        :return: A transformed input and modified transformation record
        """
        import lief # lgtm [py/repeated-import]

        random.seed(self.seed)

        binary = lief.PE.parse(list(x))

        if transformation_value is None:
            transformation_value = random.choice(list(self.transformation_range.keys()))

        # Find the library if it already exists in the import list
        lib = None
        for im in binary.imports:
            if im.name.lower() == transformation_value.lower():
                lib = im
                break

        # If not, add it
        if lib is None:
            lib = binary.add_library(transformation_value)

        # Then add a random function that doesn't exist
        possible_values = list(
            set(self.transformation_range[transformation_value]) - set([e.name for e in lib.entries])
        )
        func_to_add = random.choice(possible_values)
        lib.add_entry(func_to_add)

        new_x = binary_to_bytez(binary, imports=True)

        # update the transformation record
        transformation_record["prev_state"] = deepcopy(transformation_record["current_state"])
        if self.subtransformer_index is not None:
            transformation_record["current_state"]["actions_taken"][self.subtransformer_index] += 1
        else:
            transformation_record["current_state"]["actions_taken"] += 1

        transformation_record["current_state"]["value"] = new_x

        return new_x, transformation_record

    def get_action_list(self):
        """
        Return a list of possible library imports
        """
        action_list = list(self.transformation_range.keys())
        if self.max_actions > 0:
            random.shuffle(action_list)
            return action_list[: self.max_actions]
        return action_list

    def is_possible(self, x, transformation_value):
        """
        Determines if the action specified by transformation_value can be performed on the current input.
        :param x: Input value
        :param transformation_value: The definition of the transformation to apply

        :return: If action is possible and the transformation value. The effect on the record is not used
        """
        import lief

        binary = lief.PE.parse(list(x))

        # Find the library if it already exists in the import list
        lib = None
        for im in binary.imports:
            if im.name.lower() == transformation_value.lower():
                lib = im
                break

        # If library is not in the list, we can definitely append
        if lib is None:
            return True, [[transformation_value]], None

        # Can't append library if there are no functions left to append
        possible_values = list(
            set(self.transformation_range[transformation_value]) - set([e.name for e in lib.entries])
        )
        if len(possible_values) == 0:
            return False, None, None

        return True, [transformation_value], None

    def get_possible(self, x):
        """
        Determines what actions can be performed on the input. It randomly generates the action list and then filters out any
        libraries that are completly imported
        :param x: input_value

        :return: A list of transformation values
        """
        all_actions = self.get_action_list()
        possible_actions = []
        for action in all_actions:
            is_possible, _, _ = self.is_possible(x, action)
            if is_possible:
                possible_actions.append(action)

        return possible_actions



from uret.transformers import SubTransformer
import warnings

import numpy as np


class Increment(SubTransformer):

    name = "Increment"

    def __init__(self, low=0, high=1, number_type="float", action_samples=10, sample_method="random"):
        """
        Initialize an `Increment` object. This object increments an object by a numerical amount depending on the hyper
        parameters.
        :param low: The lower bound (inclusive) to sample increment values from. It should be a non-negative number.
        :param high: The upper bound (exclusive) to sample increment values from. It should be a non-negative number.
        :param number_type: specifies the type of number that is being modified. Currently supports "int" or "float"
        :param sample_num: Max number of samples generated when get_action_list() is called.
        :param sample_method: Determines how action samples are generated when get_action_list() is called
            - random: Generate samples at random within the range (low, high) with a uniform probability (Default)
            - linspace: Generate sample using linearly spaced samples within the range (low, high)
            - geomspace: Generate sample using log spaced samples within the range (low, high)
        """

        if abs(high) <= abs(low):
            raise ValueError("high must be greater than low")

        self.low = abs(low)
        self.high = abs(high)

        self.number_type = number_type
        if number_type == "int":
            if high - low < int(action_samples / 2):  # Corner case so the number of samples isn't more than the range
                self.action_samples = 2 * (high - low)
            else:
                self.action_samples = action_samples
        else:
            self.action_samples = action_samples
        
        if sample_method.lower() not in ["random", "linspace", "geomspace"]:
            raise ValueError("Sample method not supported. Supported methods are random,linspace, and geomspace")
        self.sample_method = sample_method.lower()

        super(Increment, self).__init__()

    def transform(
        self,
        x,
        transformation_record,
        transformation_value,
    ):
        """
        Public method. First it applies the input processing if defined, then performs the input transformation.
        :param x: The input value to be transformed
        :param transformation_record: A record tracking the transformation already applied to x. This is used for constraint
        enforcement and action validation
        :param transformation_value: A class specific argument that defines the transformation to apply to x.
        This can be a single value or a tuple/list of values depending on the implementation.

        :return: A transformed input and the new transformation record
        """

        if transformation_value is None:
            multiplier = 1 if np.random.random() < 0.5 else -1

            if self.number_type == "int":
                increment_value = multiplier * np.random.randint(self.low, self.high)
            elif self.number_type == "float":
                increment_value = multiplier * np.random.uniform(self.low, self.high)

        else:
            increment_value = transformation_value

        new_x = x + increment_value
        transformation_record = transformation_record + increment_value

        return new_x, transformation_record

    def get_action_list(self):
        """
        Return a list of possible values to increment by
        """
        
        sample_num = int(self.action_samples / 2)  # We divide by two cause we will be using pos and negative values

        low = self.low
        high = self.high

        if self.sample_method == "random":
            if self.number_type == "int":
                increment_values = np.random.randint(low, high, sample_num)
            elif self.number_type == "float":
                increment_values = np.random.uniform(low, high, sample_num)

        elif self.sample_method == "linspace":
            if self.number_type == "int":
                increment_values = np.linspace(low, high, sample_num, endpoint=False, dtype=int)
            elif self.number_type == "float":
                increment_values = np.linspace(low, high, sample_num, endpoint=False)

        elif self.sample_method == "geomspace":
            if low == 0:
                low = 0.000001
            if self.number_type == "int":
                increment_values = np.geomspace(low, high, sample_num, endpoint=False, dtype=int)
            elif self.number_type == "float":
                increment_values = np.geomspace(low, high, sample_num, endpoint=False)

        increment_values = np.array(
            list(set(increment_values))
        )  # remove duplicates. This can be from geomspace with ints usually
        increment_values = np.delete(increment_values, increment_values == 0)  # remove 0's
        
        increment_values = np.concatenate((-1 * increment_values, increment_values))  # Use positive and negative values

        return increment_values

    def is_possible(self, x, transformation_value):
        """
        Determines if the action specified by tranformation_value can be performed on the current input. If it can, it returns
        True and provides a list of the arguments to use with transform().
        :param x: Input value
        :param transformation_value: The definition of the transformation to apply

        :return: True if possible, arguments to use for transform(), and the change in delta
        """

        return True, [transformation_value], transformation_value

    def get_possible(self, x):
        """
        Determines what actions can be performed on the input. For numbers, anything is possible because the subtransformer
        doesn't need to check constraints
        :param x: Input Value

        :return: A list of transformation values
        """

        return self.get_action_list()



from uret.transformers.basic.subtransformers.increment import Increment
from uret.transformers.basic.subtransformers.category_modifier import CategoryModifier
from uret.transformers.basic.subtransformers.delete import Delete
from uret.transformers.basic.subtransformers.insert import Insert
from uret.transformers.basic.subtransformers.substitution import Substitution



from uret.transformers import SubTransformer
import warnings

import numpy as np


class CategoryModifier(SubTransformer):

    name = "CategoryModifier"

    def __init__(self, categories, is_onehot=False):
        """
        Initialize an `CategoryModifier` object.
        :param categories: A list possible feature values if not a one-hot feature. If the feature is a one-hot feature, then this  is ignored.
        :param is_onehot: Indicates if the feature is a single category featuer or a one-hot (ie multi category) feature
        """

        self.categories = categories
        self.is_onehot = is_onehot

        super(CategoryModifier, self).__init__()

    def transform(self, x, transformation_record, transformation_value):
        """
        :param x: The input value to be transformed. If a one-hot feature, then this contains all of the values
        :param transformation_record: Not used for enforcement
        :param transformation_value: Either the category to switch or the index of the category to set hot
        This can be a single value or a tuple/list of values depending on the implementation.

        :return: A transformed input and the new transformation record
        """

        if self.is_onehot and len(x) != self.categories:
            raise RuntimeError(
                f"Onehot feature of length {len(x)}, but categories is set to {self.categories}. These should be equal for onehot features"
            )

        if transformation_value is None:
            possible_values = self.get_possible()
            transformation_value = random.choice(possible_values)

        if self.is_onehot:
            x[x == 1] = 0
            x[transformation_value] = 1
        else:
            x = transformation_value
            
        transformation_record = transformation_value

        return x, transformation_record

    def get_action_list(self):
        """
        Return a list of transformation values
        """
        if self.is_onehot:
            return list(set(np.arange(self.categories)))

        return list(set(self.categories))

    def is_possible(self, x, transformation_value):
        """
        Determines if the action specified by tranformation_value can be performed on the current input. If it can, it returns
        True and provides a list of the arguments to use with transform().
        :param x: Input value
        :param transformation_value: The index to transform (one-hot) or category to change to

        :return: True if possible, arguments to use for transform()
        """

        if self.is_onehot and x[transformation_value] == 1:
            return False, [], None
        elif not self.is_onehot and x == transformation_value:
            return False, [], None
        return True, [transformation_value], transformation_value

    def get_possible(self, x):
        """
        Determines what actions can be performed on the input.
        :param x: Input Value

        :return: A list of transformation values
        """

        if self.is_onehot:
            if len(x) != self.categories:
                raise RuntimeError(
                    f"Onehot feature of length {len(x)}, but categories is set to {self.categories}. These should be equal for onehot features"
                )
            possible_values = list(set(np.arange(self.categories)) - set(np.where(x == 1)[0]))
        else:
            possible_values = list(set(self.categories) - set(list(x)))

        return possible_values



from uret.transformers import SubTransformer
import string

import numpy as np
from copy import deepcopy


class Insert(SubTransformer):

    name = "Insert"

    def __init__(self, transformation_range=None, subtransformer_index=None):
        """
        Initialize an `Insert` object. This object inserts a character into a string.
        :param transformation_range: The characters that can be inserted.
        :param subtransformer_index: The index of the subtransformer in the transformer list. If not none, the subtransformer assumes the transformation record 'actions_taken' value is a list.
        """
        if transformation_range is None:
            self.transformation_range = string.ascii_lowercase + string.digits
        else:
            self.transformation_range = transformation_range

        self.subtransformer_index = subtransformer_index
        super(Insert, self).__init__()

    def transform(self, x, transformation_record, transformation_value):
        """
        Inserts a value from to the input x
        :param x: Input text
        :param transformation_record: A record tracking the transformation already applied to x. This is used for constraint
        enforcement and action validation
        :param transformation_value: Target character to insert

        :return: A transformed input
        :rtype:
        """

        if transformation_value is None:
            transformation_value = np.random.choice(self.transformation_range)
        elif transformation_value not in self.transformation_range:
            raise ValueError(transformation_value + " not in range")

        replacement_pos = np.random.randint(len(x))

        new_x = x[:replacement_pos] + transformation_value + x[replacement_pos:]

        transformation_record["prev_state"] = deepcopy(transformation_record["current_state"])
        if self.subtransformer_index is not None:
            transformation_record["current_state"]["actions_taken"][self.subtransformer_index] += 1
        else:
            transformation_record["current_state"]["actions_taken"] += 1

        transformation_record["current_state"]["delta"] = transformation_record["current_state"]["delta"] + len(
            transformation_value
        )
        transformation_record["current_state"]["value"] = new_x

        return new_x, transformation_record

    def get_action_list(self):
        """
        Return a list of characters that the subtransformer can insert
        """
        action_list = list(self.transformation_range)
        return action_list

    def is_possible(self, x, transformation_value):
        """
        Determines if the action specified by transformation_value can be performed on the current input. If it can, it returns
        True and provides a list of the arguments to use with transform().
        :param x: Input value
        :param transformation_value: The definition of the transformation to apply

        :return: True if possible, arguments to use for transform(), and the change in string length
        """
        if transformation_value not in self.transformation_range:
            return False, None, 0

        return True, [transformation_value], len(transformation_value)

    def get_possible(self, x):
        """
        Determines what actions can be performed on the input. For string, only characters in the string range can be inserted
        :param x: Input Value

        :return: A list of transformation values
        """

        return self.get_action_list()



from uret.core.explorers.graph_explorer import GraphExplorer
from uret.core.rankers import Random
import simanneal
import warnings

import copy
import random


class SimulatedAnnealingSearchGraphExplorer(GraphExplorer, simanneal.Annealer):
    """
    This class implement simulated annealing (random walk based on an energy function) using `simanneal` package.
    """

    def __init__(
        self,
        model_predict,
        ranking_algorithm,
        feature_extractor=None,
        scoring_alg="model_loss",
        scoring_function=None,
        dependencies=None,
        target_label=None,
        attack_time=0.5,
        restarts=0,
        schedule=None,
        schedule_steps=100,
        min_transform_c_sampled=1,
        min_transform_i_sampled=1,
        global_max_transforms=None,
        max_transform_c_sampled=None,
        max_transform_i_sampled=None,
        repeat_probability = 0.5,
        verbose=False,
    ):
        """
        Create a graph explorer that uses the beam search algorithm to generate adversarial samples

        :param model_predict: Classification model that the adversarial sample is being generated for. This should just be the callable prediction function of the model. It is assume it outputs a 1-hot vector or a class probability vector
        :param ranking_algorithm: Algorithm used to estimate the value of the neighboring vertices
        :param feature_extractor: Algorithm that converts a raw input into a input that the model can ingest
        :param scoring_alg: What type of loss will be used to score a vertex. This informs the explorer what parameters should be passed to the scoring function
        :param scoring_function: Function used to score a vertex (i.e. input sample). This should take in 2 or 3 arguments depending on the loss type and return a value in which, the lower the value, the "better" the index at achieving the goal
        :param dependencies: A list of functions in the form [callable, callable_args] that enforce input dependencies (e.g. features 1 and 2 should always add up to feature 3)
        :param target_label: Set this value to a class index in the model's output range to perform a targeted attack. If none, it is assume that we are performing an untargeted attack by default.
        :param attack_time: Time to spend to perform annealing. Note that if auto scheduling is used, the initial schedule search will be longer than the attack time.
        :param restarts: Number of times to perform annealing
        :param schedule: Annealing specific param. Can be provided by the user if auto schedule is not desired
        :param schedule_steps: If autoscheduling is used, this is the number of steps.
        :param min_transform_c_sampled: Min number of transformers that use the current state to randomly select
        :param min_transform_i_sampled: Min number of transformers that use the initial state to randomly select
        :param global_max_transforms: Max number of transforms to apply regardless of state. For multi-feature inputs, this will trigger a state reset if the randomly selected transforms will modifying more features than global_max_transforms. For single inputs, this will trigger a reset after move() has been called global_max_transforms times.
        :param max_transform_c_sampled: Max number of transformers that use the current state to randomly select
        :param max_transform_i_sampled: Max number of transformers that use the initial state to randomly select
        :param repeat_probability: If a transformation is to be repeated and it doesn't have a max_action constraint, flip a 
        weighted coin until failures to chose a number of repeats.
        :param verbose: disable simanneal output if false
        """

        if not isinstance(ranking_algorithm, Random):
            raise ValueError(f"The ranking algorithm must be Random")

        super().__init__(
            model_predict=model_predict,
            ranking_algorithm=ranking_algorithm,
            feature_extractor=feature_extractor,
            scoring_alg=scoring_alg,
            scoring_function=scoring_function,
            dependencies=dependencies,
            target_label=target_label,
        )

        # Get the inds for tranformer based on whether they transform the intial or current state
        # Also track which transformers must be called multiple times (e.g. string transformers)
        self.transform_initial_inds = []
        self.transform_current_inds = []
        self.transform_repeat_inds = []
        for transformer_index, (transformer, _) in enumerate(self.ranking_algorithm.transformer_list):
            if transformer.transform_initial:
                self.transform_initial_inds.append(transformer_index)
            else:
                self.transform_current_inds.append(transformer_index)

            if transformer.transform_repeat:
                self.transform_repeat_inds.append(transformer_index)

        # Simanneal params
        self.attack_time = attack_time
        self.restarts = restarts
        self.schedule = schedule  # This is automatically tuned
        self.schedule_steps = schedule_steps  # This is automatically tuned

        # Sampling Params
        self.global_max_transforms = global_max_transforms
        self.move_calls = 0
        self.min_transform_i_sampled = min(min_transform_i_sampled, len(self.transform_initial_inds))
        self.max_transform_i_sampled = (
            max_transform_i_sampled
            if max_transform_i_sampled and max_transform_i_sampled <= len(self.transform_initial_inds)
            else len(self.transform_initial_inds)
        )

        self.min_transform_c_sampled = min(min_transform_c_sampled, len(self.transform_current_inds))
        self.max_transform_c_sampled = (
            max_transform_c_sampled
            if max_transform_c_sampled and max_transform_c_sampled <= len(self.transform_current_inds)
            else len(self.transform_current_inds)
        )

        # Catch bad values of max
        if (
            self.max_transform_i_sampled < self.min_transform_i_sampled
            or self.max_transform_c_sampled < self.min_transform_c_sampled
        ):
            warnings.warn(
                f"A max transform value was less than the min transforms. Automatically setting it equal to min transforms. Be careful next time!"
            )
            self.max_transform_i_sampled = max(self.max_transform_i_sampled, self.min_transform_i_sampled)
            self.max_transform_c_sampled = max(self.max_transform_c_sampled, self.min_transform_c_sampled)

        self.repeat_prob = repeat_probability
        self.verbose = verbose

    def move(self):
        """
        Auxiliary function used by `anneal` function. This function randomly changes `self.state` using the transformations.
        """
        transform_current_inds = []
        transform_initial_inds = []

        # Get the transformers to use
        if len(self.transform_current_inds) > 0:
            transform_current_inds = random.sample(
                self.transform_current_inds,
                k=random.randint(self.min_transform_c_sampled, self.max_transform_c_sampled),
            )

        if len(self.transform_initial_inds) > 0:
            transform_initial_inds = random.sample(
                self.transform_initial_inds,
                k=random.randint(self.min_transform_i_sampled, self.max_transform_i_sampled),
            )

        all_inds = transform_current_inds + transform_initial_inds

        
        if self.global_max_transforms:
            all_inds = random.sample(all_inds, k=min(len(all_inds), self.global_max_transforms))
            # Reset the state for multi feature inputs if the new transformations would go past the global max
            if self.ranking_algorithm.multi_feature_input:
                prev_modified_inds = [i for i, tr in enumerate(self.state["transformation_records"]) if tr is not None]
                to_be_modified_inds = set(all_inds).union(prev_modified_inds)
                if len(to_be_modified_inds) > self.global_max_transforms:
                    self.state = {"sample": copy.deepcopy(self.initial_sample_state), "transformation_records": copy.deepcopy(self.initial_record_state)}
            # Reset the state for single inputs if the new transformations would go past the global max
            elif not self.ranking_algorithm.multi_feature_input:
                if self.move_calls >= self.global_max_transforms:
                    self.state = {"sample": copy.deepcopy(self.initial_sample_state), "transformation_records": copy.deepcopy(self.initial_record_state)}
                    self.move_calls = 0
                self.move_calls += 1
                    
                

        # Transform
        for i in all_inds:
            # Check if the transformer must be repeated. If so, repeat some number of times.
            if i in self.transform_repeat_inds:
                if "max_actions" in self.ranking_algorithm.transformer_list[i][0].input_constraints.keys():
                    max_actions = self.ranking_algorithm.transformer_list[i][0].input_constraints["max_actions"]
                    repeat_transform = random.randint(1, max_actions)
                # Otherwise, just repeat some number of times by flipping a coin
                else:
                    repeat_transform = 1
                    while random.random() < self.repeat_prob:
                        repeat_transform += 1

            else:
                repeat_transform = 1

            # Get the intial transform state
            if i in self.transform_initial_inds:
                transformed_state = copy.deepcopy(self.initial_sample_state)
                transformation_record = None
            else:
                transformed_state = copy.deepcopy(self.state["sample"])
                transformation_record = copy.deepcopy(self.state["transformation_records"])

            # Transform
            while repeat_transform > 0:
                return_values = self.ranking_algorithm.rank_edges(
                    transformed_state,
                    self.scoring_function,
                    self.score_input,
                    dependencies=self.dependencies,
                    current_transformation_records=transformation_record,
                    transformer_index=i,
                )
                indicies, _, _, transformed_state, transformation_record, _ = return_values[
                    0
                ]  # We only expect 1 return value
                repeat_transform -= 1

            if self.ranking_algorithm.multi_feature_input:
                self.state["sample"][indicies[1]] = transformed_state[indicies[1]]
                self.state["transformation_records"][indicies[0]] = transformation_record[indicies[0]]
            else:
                self.state["sample"] = transformed_state
                self.state["transformation_records"] = transformation_record

    def search(self, sample, score_input):
        """
        Perform simulated annealing and return the lowest energy sample
        """

        convert_back_to_list = False
        if self.ranking_algorithm.multi_feature_input and isinstance(sample, list):
            convert_back_to_list = True
            sample = np.array(sample)

        # Create transformation record
        init_transformation_records = None
        if self.ranking_algorithm.multi_feature_input:
            init_transformation_records = [None for _ in range(len(self.ranking_algorithm.transformer_list))]

        self.initial_sample_state = copy.deepcopy(sample)
        self.initial_record_state = copy.deepcopy(init_transformation_records)
        self.state = {"sample": copy.deepcopy(self.initial_sample_state), "transformation_records": copy.deepcopy(self.initial_record_state)}
        self.score_input = score_input

        if self.schedule is None:
            self.schedule = self.auto(minutes=self.attack_time, steps=self.schedule_steps)
            self.set_schedule(self.schedule)

        for i in range(self.restarts + 1):
            self.initial_sample_state = copy.deepcopy(sample)
            self.initial_record_state = copy.deepcopy(init_transformation_records)
            self.state = {"sample": copy.deepcopy(self.initial_sample_state), "transformation_records": copy.deepcopy(self.initial_record_state)}
            state, energy = self.anneal()
            if convert_back_to_list:
                yield list(state["sample"]), state["transformation_records"], energy
            else:
                yield state["sample"], state["transformation_records"], energy

    def energy(self):
        """
        returns the energy to be minimized. This function is used by `simanneal` package in functions such as `self.auto` and `self.anneal`.
        """
        return self.scoring_function(self.state["sample"], self.score_input)

    def update(self, *args, **kwargs):
        """
        Wrapper for internal update. Prints out current annealing values
        If you override the self.update method,
        you can chose to call the self.default_update method
        from your own Annealer.
        """
        if self.verbose:
            self.default_update(*args, **kwargs)



from uret.core.rankers.ranking_algorithm import RankingAlgorithm
import warnings
import copy


class BruteForce(RankingAlgorithm):
    """
    This implementation tries all transformations and parameters for each given sample,
    and returns the list of them with scores and applied transformations.
    """

    def __init__(self, transformer_list, multi_feature_input=False, is_trained=True):
        """
        :param transformer_list: List of available transformations to be used by the algorithm.
        :param multi_feature_input: A boolean flag indicating if each index in the transformer
            list defines a transformer(s) for a single data type in the input. If False, then it
            is assumed that the transformer list defines multiple transformers for a single input
        :param is_trained: Indicates when _train has been called at least once. Always True
        """
        super().__init__(transformer_list, multi_feature_input, is_trained=True)

    def rank_edges(self, sample, scoring_function, score_input, dependencies=[], current_transformation_records=None):

        # Create transformation record
        if self.multi_feature_input and current_transformation_records is None:
            current_transformation_records = [None for _ in range(len(self.transformer_list))]

        return_values = []  # This will contain the (sample_index, transformer, action args, transformed sample,
        # transformation_record of the transformed sample, score)
        
        for transformer_index, (transformer, input_index) in enumerate(self.transformer_list):
            if self.multi_feature_input:
                possible_actions = transformer.get_possible(
                    sample[input_index], transformation_record=current_transformation_records[transformer_index]
                )
            else:
                possible_actions = transformer.get_possible(
                    sample, transformation_record=current_transformation_records
                )
            for action in possible_actions:
                sample_temp = copy.copy(sample)
                transformation_records_temp = copy.deepcopy(current_transformation_records)

                if self.multi_feature_input:
                    transformed_value, new_transformation_record = transformer.transform(
                        sample_temp[input_index],
                        transformation_record=transformation_records_temp[transformer_index],
                        transformation_value=action,
                    )
                    sample_temp[input_index] = transformed_value
                    transformation_records_temp[transformer_index] = new_transformation_record
                else:
                    transformed_value, new_transformation_record = transformer.transform(
                        sample_temp, transformation_record=transformation_records_temp, transformation_value=action
                    )
                    sample_temp = transformed_value
                    transformation_records_temp = new_transformation_record

                sample_temp = self._enforce_dependencies(sample_temp, dependencies)
                score = scoring_function(sample_temp, score_input)

                if self.multi_feature_input:
                    return_values.append(
                        (
                            [transformer_index, input_index],
                            transformer,
                            action,
                            sample_temp,
                            transformation_records_temp,
                            score,
                        )
                    )
                else:
                    return_values.append([None, transformer, action, sample_temp, transformation_records_temp, score])
        return return_values



from abc import ABC, abstractmethod


class SubTransformer(ABC):
    """
    The abstract base class `SubTransformer` defines the basic requirements of a SubTransformer. A SubTransformer is a class that manipulates a specified data type.
    """

    def __init__(self):
        """
        Initialize a `SubTransformer` object.
        """
        pass

    @abstractmethod
    def transform(self, x, transformation_record, transformation_value=None, *args, **kwargs):
        """
        Public method. First it applies the input processing if defined, then performs the input transformation.
        :param x: The input value to be transformed
        :param transformation_record: A record tracking the transformation already applied to x. This is used for constraint
        enforcement and action validation
        :param transformation_value: The definition of the transformation to apply

        :return: A transformed input and the new transformation record
        """
        raise NotImplementedError

    @abstractmethod
    def get_action_list(self):
        """
        Return a list of possible actions for the transformation
        """
        raise NotImplementedError

    @abstractmethod
    def is_possible(self, x, transformation_value):
        """
        Determines if the action specified by tranformation_value can be performed on the current input. If it can, it returns
        True and provides a list of the arguments to use with transform().
        :param x: Input value
        :param transformation_value: The definition of the transformation to apply

        :return: True if possible, arguments to use for transform(), and effect on the transformation record (from Transformer)
        with respect to the Transformer's possible input constraints
        """
        raise NotImplementedError

    @abstractmethod
    def get_possible(self, x):
        """
        Determines what actions can be performed on the input.
        :param x: Input Value

        :return: A list of transformation values
        """
        raise NotImplementedError



"""
Universal Robustness Evaluation Toolkit (URET).
"""
import logging.config

# Project Imports
from uret import core
from uret import transformers
from uret import utils

# Semantic Version
__version__ = "0.1"


from uret.transformers.basic.number_transformer import NumberTransformer
from uret.transformers.basic.string_transformer import StringTransformer
from uret.transformers.basic.category_transformer import CategoryTransformer



from uret.transformers import SubTransformer
import random
import re
import string

from copy import deepcopy


class Delete(SubTransformer):

    name = "Delete"

    def __init__(self, transformation_range=None, subtransformer_index=None):
        """
        Initialize a `Delete` object. This object deletes a character from a string.
        :param transformation_range: The characters that can be deleted.
        :param subtransformer_index: The index of the subtrasformer in the transformer list. If not none, the subtransformer assumes the transformation record 'actions_taken' value is a list.
        """
        if transformation_range is None:
            self.transformation_range = string.ascii_lowercase + string.digits
        else:
            self.transformation_range = transformation_range

        self.subtransformer_index = subtransformer_index
        super(Delete, self).__init__()

    def transform(self, x, transformation_record, transformation_value, index=None):
        """
        Remove a value from to the input x
        :param x: Input text
        :param transformation_record: A record tracking the transformation already applied to x. This is used for constraint
        enforcement and action validation
        :param transformation_value: Target character to remove
        :param index: Index of the character to remove

        :return: A transformed input
        :rtype:
        """

        if transformation_value is None:
            modifiable_values = [v for v in x if v in self.transformation_range]
            transformation_value = random.choice(modifiable_values)

        if index is None:
            replace_inds = [match.start() for match in re.finditer(transformation_value, x)]
            replacement_pos = random.choice(replace_inds)
            if len(replace_inds) == 0:
                print(transformation_value, "not present")
                return x, transformation_record

        elif x[index] == transformation_value:
            replacement_pos = index
        else:
            raise ValueError(transformation_value + " not at index")

        new_x = x[:replacement_pos] + x[replacement_pos + len(transformation_value) :]

        transformation_record["prev_state"] = deepcopy(transformation_record["current_state"])

        if self.subtransformer_index is not None:
            transformation_record["current_state"]["actions_taken"][self.subtransformer_index] += 1
        else:
            transformation_record["current_state"]["actions_taken"] += 1

        transformation_record["current_state"]["delta"] = transformation_record["current_state"]["delta"] - 1 * len(
            transformation_value
        )
        transformation_record["current_state"]["value"] = new_x

        return new_x, transformation_record

    def get_action_list(self):
        """
        Return a list of characters that the subtransformer can delete
        """
        action_list = list(self.transformation_range)
        return action_list

    def is_possible(self, x, transformation_value):
        """
        Determines if the action specified by transformation_value can be performed on the current input. If it can, it returns
        True and provides a list of the arguments to use with transform().
        :param x: Input value
        :param transformation_value: The definition of the transformation to apply

        :return: True if possible, arguments to use for transform(), and the change in string length
        """
        if transformation_value not in self.transformation_range:
            return False, None, 0

        replace_inds = [match.start() for match in re.finditer(transformation_value, x)]

        if len(replace_inds) == 0:
            return False, None, 0

        return True, [[transformation_value, ind] for ind in replace_inds], -1 * len(transformation_value)

    def get_possible(self, x):
        """
        Determines what actions can be performed on the input. For string, only characters in the string can be deleted
        :param x: Input Value

        :return: A list of transformation values
        """

        uniq_chars = [c for c in list(set(x)) if c in self.transformation_range]

        return uniq_chars



import codecs
import os

from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

install_requires = [
    "numpy>=1.18.0",
    "scipy>=1.4.1",
    "scikit-learn>=0.22.2",
    "setuptools",
    "tqdm",
    "simanneal",
]


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r", encoding="utf-8") as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")


setup(
    name="uret",
    version=get_version("uret/__init__.py"),
    description="Toolkit for generic adversarial machine learning evaluationsI.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Kevin Eykholt",
    author_email="kheykholt@ibm.com",
    maintainer="Kevin Eykholt",
    maintainer_email="kheykholt@ibm.com",
    url="https://github.com/IBM/URET",
    license="MIT",
    install_requires=install_requires,
    extras_require={
        "all": ["lief", "pandas", "tensorflow", "keras", "h5py", "keras-rl"],
        "binary": ["lief"],
        "rl": ["tensorflow", "keras", "h5py", "keras-rl"],
        "non-framework": ["pandas"]
    },
    packages=find_packages(),
    include_package_data=True,
)


from uret.core.explorers.graph_explorer import GraphExplorer
import numpy as np
import copy


class BeamSearchGraphExplorer(GraphExplorer):
    """
    This class implement Beam Search using recursion.
    """

    def __init__(
        self,
        model_predict,
        ranking_algorithm,
        feature_extractor=None,
        scoring_alg="model_loss",
        scoring_function=None,
        dependencies=None,
        target_label=None,
        search_size=3,
        max_depth=2,
        max_visits=None,
    ):
        """
        Create a graph explorer that uses the beam search algorithm to generate adversarial samples

        param model_predict: Classification model that the adversarial sample is being generated for. This should just be the callable prediction function of the model. It is assume it outputs a 1-hot vector or a class probability vector
        param ranking_algorithm: Algorithm used to estimate the value of the neighboring vertices
        param feature_extractor: Algorithm that converts a raw input into a input that the model can ingest
        param scoring_alg: What type of loss will be used to score a vertex. This informs the explorer what parameters should be passed to the scoring function
        param scoring_function: Function used to score a vertex (i.e. input sample). This should take in 2 or 3 arguments depending on the loss type and return a value in which, the lower the value, the "better" the index at achieving the goal
        param dependencies: A list of functions in the form [callable, callable_args] that enforce input dependencies (e.g. features 1 and 2 should always add up to feature 3)
        param target_label: Set this value to a class index in the model's output range to perform a targeted attack. If none, it is assume that we are performing an untargeted attack by default.
        param search_size: How many top neighbors to explore at each step/for each node.
        param max_depth: How many steps to explore, or how many times we stack `neighbor_algorithm` on a sample.
        max_visits: The max number of generated samples to visit.
        """

        super().__init__(
            model_predict=model_predict,
            ranking_algorithm=ranking_algorithm,
            feature_extractor=feature_extractor,
            scoring_alg=scoring_alg,
            scoring_function=scoring_function,
            dependencies=dependencies,
            target_label=target_label,
        )

        self.search_size = search_size
        self.max_depth = max_depth
        self.visited_nodes = []  # Keep track of visited nodes to avoid revisits
        self.max_visits = max_visits

    def search(self, sample, score_input, transformation_records=None, depth=0):
        """
         runs search and returns (sample candidate, list of applied transformations, score).
         This method implements beam sesarch in a depth-first and recursive manner, taking top-`self.search_size` samples,
         and repeating the function up to `self.max_depth` times.

         :param sample: an input sample, which can be either the raw input for the feature extractor,
                        or the feature vector produced by the feature extractor that will be fed to the model.
        :param score_input: Input value for the scoring function. This is either the original pred or the target feature value
        :param transformation_records: history of applied transformations.
        :param depth: current depth of search.
        """
        if depth >= self.max_depth:
            return
        if depth == 0:
            self.visited_nodes = []

        convert_back_to_list = False
        if self.ranking_algorithm.multi_feature_input and isinstance(sample, list):
            convert_back_to_list = True
            sample = np.array(sample)

        edge_transform_estimates = self.ranking_algorithm.rank_edges(
            sample, self.scoring_function, score_input, self.dependencies, transformation_records
        )

        # No actions are possible from current vertex
        if len(edge_transform_estimates) == 0:
            return

        indicies, transformers, transformer_params, samples_next, new_transformation_records, scores = zip(
            *edge_transform_estimates
        )

        ranked_edge_list = np.argsort(scores)  # Score is the lower, the better.
        rank = 0
        returned_edge_count = 0
        while returned_edge_count < self.search_size and rank < len(ranked_edge_list):
            if rank >= len(ranked_edge_list):
                break
            edge_idx = ranked_edge_list[rank]
            # Get the values for the edge that is being evaluated
            indicies_next = indicies[
                edge_idx
            ]  # if not none, this indicates which feature of x the transformer transforms
            transformer_next = transformers[edge_idx]
            transformer_param_next = transformer_params[edge_idx]
            sample_next = samples_next[edge_idx]
            transformation_records_next = new_transformation_records[edge_idx]

            # Apply the transformation, if not done by the neighbor eval algorithm.
            if sample_next is None:
                if indicies_next is not None:
                    sample_next = copy.copy(sample)
                    transformation_records_next = copy.copy(transformation_records)
                    if self.ranking_algorithm.multi_feature_input and transformation_records_next is None:
                        transformation_records_next = [
                            None for _ in range(len(self.ranking_algorithm.transformer_list))
                        ]

                    transformed_value, new_transformation_record = transformer_next.transform(
                        sample_next[indicies_next[1]],
                        transformation_record=transformation_records_next[indicies_next[0]],
                        transformation_value=transformer_param_next,
                    )
                    sample_next[indicies_next[1]] = transformed_value
                    transformation_records_next[indicies_next[0]] = new_transformation_record
                else:
                    sample_next, transformation_records_next = transformer_next.transform(
                        sample,
                        transformation_record=transformation_records,
                        transformation_value=transformer_param_next,
                    )

                sample_next = self._enforce_dependencies(
                    sample_next
                )  # Enforce dependencies since ranking algorithm did not

            # Only evaluate nodes that haven't been previously visited
            if not np.any(
                [np.all(sample_next == v) for v in self.visited_nodes]
            ):  # This might not work with all data types?
                if convert_back_to_list:  # Restore back to list
                    sample_next = list(sample_next)

                self.visited_nodes.append(sample_next)
                score_next = scores[edge_idx]
                yield sample_next, transformation_records_next, score_next
                returned_edge_count += 1
                if self.max_visits and len(self.visited_nodes) >= self.max_visits:
                    break
                for sts_tuple in self.search(
                    sample_next,
                    score_input,
                    transformation_records_next,
                    depth + 1,
                ):  # recursion
                    yield sts_tuple
            rank += 1


def GreedySearchGraphExplorer(
    model_predict,
    ranking_algorithm,
    feature_extractor=None,
    scoring_alg="model_loss",
    scoring_function=None,
    dependencies=[],
    target_label=None,
    max_depth=2,
):
    """
    Alias for BeamSearchGraphExplorer with search size 1, always taking the best option greedily.
    """
    return BeamSearchGraphExplorer(
        model_predict,
        ranking_algorithm,
        feature_extractor,
        scoring_alg,
        scoring_function,
        dependencies,
        target_label,
        search_size=1,
        max_depth=max_depth,
    )



from abc import ABC, abstractmethod
import importlib  # This is used to import the subtransformers from a fixed directory
import random


class Transformer(ABC):
    """
    The abstract base class `Transformer` defines the basic requirements of a transformer
    """

    # Two attributes used by simulated annealing to track transformer properties
    transform_initial = False  # If True, then inputs to transform() should be based on the initial state of the sample
    transform_repeat = (
        False  # If True, then transform() should be called multiple times during one annealing move() call
    )

    def __init__(self, subtransformer_args, input_constraints={}, input_processor=None):
        """
        Initialize a `Transformer` object.
        :param subtransformer_args: Arguments used to initialize the subtransformers, which are functions that manipulate the
        input data. For example, if the Transformer was defined for numerical vlues, then a subtransformer might be a function
        that increments the value.
        :param input_constraints: The constraints to enforce on the input after transformation. For example, for numerical values
        these constraints might be clip bounds. The input constraints specific to the transformer will decide the structure of the
        'transformation_record'
        :param input_processor: A function with two modes to control transforming input values where parts of the input should
        remain unchanged. If a single input is given, the function will return (the modifiable sections, unmodifiable sections).
        If two inputs are given, it is assume they are in the order (the modifiable sections, unmodifiable sections) and the
        function will fuse them together. For example, maybe the input is a string "abc.com", but you only want to perform
        transformations on "abc"
        """
        self.subtransformer_list = []

        # The subtransformers should be stored in <current_module_path>.subtransformers for a particular task.
        subtransformer_directory = ".".join(self.__module__.split(".")[:-1])
        subtransformer_directory = subtransformer_directory + ".subtransformers"

        # DANGER DANGER BECAUSE USER DEFINE THE INPUT VALUE. LOOK AT LATER
        for args in subtransformer_args:
            subtransformer_module = importlib.import_module(subtransformer_directory)
            subtransformer = getattr(subtransformer_module, args["name"])
            if "init_args" in args.keys():
                self.subtransformer_list.append(subtransformer(**args["init_args"]))
            else:
                self.subtransformer_list.append(subtransformer())

        self.input_constraints = input_constraints
        self.input_processor = input_processor

    @abstractmethod
    def init_transformation_record(self, x):
        """
        Define the structure of the transformation record and initialize it.
        :param x: The input value to be transformed

        :return: A record tracking the transformation already applied to x. This is used for constraint
        enforcement and action validation. The definition of the transformation record depends on the Transformer's possible input
        constraints
        """
        raise NotImplementedError

    def transform(self, x, transformation_record=None, transformation_value=None):
        """
        First it applies the input processing if defined, then performs the input transformation.
        :param x: The input value to be transformed
        :param transformation_record: A record tracking the transformation already applied to x. This is used for constraint
        enforcement and action validation. The definition of the transformation record depends on the Transformer's possible input
        constraints
        :param transformation_value: A class specific argument that defines the transformation to apply to x.
        This can be a single value or a tuple/list of values depending on the implementation.

        :return: A transformed input and the new transformation record
        """

        if self.input_processor:
            x, static_fields = self.input_processor(x)

        if transformation_record is None:
            transformation_record = self.init_transformation_record(x)

        # If nothing defined, randomly select a transformation function
        if transformation_value is None:
            transformation_index = random.choice(range(len(self.subtransformer_list)))
            x, transformation_record = self.subtransformer_list[transformation_index].transform(
                x, transformation_record, None
            )
        # Otherwise, use the specified transformation
        else:
            transformation_index = transformation_value[0]
            x, transformation_record = self.subtransformer_list[transformation_index].transform(
                x, transformation_record, *transformation_value[1:]
            )

        if len(self.input_constraints.keys()) != 0:
            x, transformation_record = self._enforce_constraints(x, transformation_record)

        if self.input_processor is not None:
            x = self.input_processor(x, static_fields)
        return x, transformation_record

    @abstractmethod
    def _enforce_constraints(self, x, transformation_record):
        """
        Enforce predefined input constraints
        :param x: Input value to be modified
        :param transformation_record: A record tracking the transformation already applied to x. This is used for constraint
        enforcement and action validation. The definition of the transformation record depends on the Transformer's possible input
        constraints

        :return: A "clipped" input and corrected trasnformation record
        """
        raise NotImplementedError

    def get_action_list(self):
        """
        Return a list of actions for the transformation specified based on the subtransformations
        """
        action_list = []
        for i in range(len(self.subtransformer_list)):
            subtransformer_action_list = self.subtransformer_list[i].get_action_list()
            for action in subtransformer_action_list:
                action_list.append([i, action])

        return action_list

    @abstractmethod
    def _is_possible(self, x, transformation_record=None, transformation_value=None):
        """
        Determines if the action specified by tranformation_value can be performed on the current input. If it can, it returns
        True and provides a list of the arguments to use with transform().
        :param x: Input value
        :param transformation_record: A record tracking the transformation already applied to x. This is used for constraint
        enforcement and action validation. The definition of the transformation record depends on the Transformer's possible input
        constraints
        :param transformation_value: The definition of the transformation to apply

        :return: True if possible, arguments to use for transform()
        """
        raise NotImplementedError

    def is_possible(self, x, transformation_record=None, transformation_value=None):

        if transformation_value is None:
            return False, []

        if transformation_record is None:
            transformation_record = self.init_transformation_record(x)

        if self.input_processor:
            x, static_fields = self.input_processor(x)

        transformation_index = transformation_value[0]
        possible, action_args = self._is_possible(
            x, transformation_record=transformation_record, transformation_value=transformation_value
        )

        if not possible:
            return False, []

        # Check if the action args contain a single argument group or a list of multiple argument groups
        if not isinstance(action_args[0], list):
            return possible, [[transformation_index, *action_args]]
        else:
            return possible, [[transformation_index, *a] for a in action_args]

    def get_possible(self, x, transformation_record=None):
        """
        Determines what actions can be performed on the input.
        :param x: Input Value
        :param transformation_record: A record tracking the transformation already applied to x. This is used for constraint
        enforcement and action validation. The definition of the transformation record depends on the Transformer's possible input
        constraints

        :return: A list of transformation values and the subtransform indices
        """
        # We do this here so if_possible doesn't need to do it every time
        if transformation_record is None:
            transformation_record = self.init_transformation_record(x)

        action_list = []
        for i, subtransformer in enumerate(self.subtransformer_list):
            subtransformer_actions = subtransformer.get_possible(x)
            for action in subtransformer_actions:
                possible, action_args = self.is_possible(
                    x, transformation_record=transformation_record, transformation_value=[i, action]
                )
                if possible:
                    for a in action_args:
                        action_list.append(a)

        return action_list



from uret.transformers import SubTransformer

import random
import os
import tempfile
import subprocess

import numpy as np
from copy import deepcopy


class UPXPack(SubTransformer):
    name = "UPXPack"

    def __init__(self, compression_levels=list(range(1, 10)), seed=None, subtransformer_index=None):
        """
        Initialize a `UPXPack` object. This object compresses the binary with UPX.
        :param compression_levels: An int or list of ints between [1,9] for the different UPX values.
        :param seed: Random seed value
        :param subtransformer_index: An int indicating where in the transformation record the subtransformer is defined. This is used to enforce "max # of transformations for this subtransformer" constraint.
        """

        if isinstance(compression_levels, int):
            self.compression_levels = [compression_levels]
        else:
            self.compression_levels = compression_levels

        if any(np.array(compression_levels) < 1) or any(np.array(compression_levels) > 9):
            raise ValueError("compression_levels must have values in the range [1,9]")

        self.seed = seed
        self.subtransformer_index = subtransformer_index

        super(UPXPack, self).__init__()

    def transform(self, x, transformation_record, transformation_value):
        """
        Pack x using UPX
        :param x: Input Value
        :param transformation_record: Record of changes
        :param transformation_value: The compression level
        :return: A transformed input and modified transformation record
        """

        random.seed(self.seed)
        tempfilename = os.path.join(tempfile._get_default_tempdir(), next(tempfile._get_candidate_names()))

        # dump to a temporary file
        with open(tempfilename, "wb") as outfile:
            outfile.write(x)

        options = ["--force", "--overlay=copy"]
        options += ["-{}".format(transformation_value)]

        # optional arguments - We'll choose at random for simplicity
        options += ["--compress-exports={}".format(random.randint(0, 1))]
        options += ["--compress-icons={}".format(random.randint(0, 3))]
        options += ["--compress-resources={}".format(random.randint(0, 1))]
        options += ["--strip-relocs={}".format(random.randint(0, 1))]

        with open(os.devnull, "w") as DEVNULL:
            retcode = subprocess.call(
                ["upx"] + options + [tempfilename, "-o", tempfilename + "_packed"], stdout=DEVNULL, stderr=DEVNULL
            )

        os.unlink(tempfilename)

        # Op succcessful
        if retcode == 0:
            with open(tempfilename + "_packed", "rb") as infile:
                x = infile.read()
            os.unlink(tempfilename + "_packed")

            # update the transformation record
            # There is a chance it is none if we are just calling this function to see if UPX exists
            if transformation_record is not None:
                transformation_record["prev_state"] = deepcopy(transformation_record["current_state"])
                if self.subtransformer_index is not None:
                    transformation_record["current_state"]["actions_taken"][self.subtransformer_index] += 1
                else:
                    transformation_record["current_state"]["actions_taken"] += 1

                transformation_record["current_state"]["value"] = x

        return x, transformation_record

    def get_action_list(self):
        """
        Return  compression levels
        """
        return self.compression_levels

    def is_possible(self, x, transformation_value):
        """
        Determines if the action specified by transformation_value can be performed on the current input.
        :param x: Input value
        :param transformation_value: The definition of the transformation to apply

        :return: If action is possible and the transformation value. The effect on the record is not used
        """

        # not sure how to check without actually packing
        try:
            self.transform(x, None, transformation_value)
            return True, [transformation_value], None
        except FileNotFoundError:
            return False, [], None  # not sure how to check without actually packing

    def get_possible(self, x):
        """
        Determines what actions can be performed on the input.
        :param x: input_value

        :return: A list of transformation values
        """
        # not sure how to check without actually packing
        try:
            self.transform(x, None, self.compression_levels[0])
            return self.compression_levels
        except FileNotFoundError:
            return []



from abc import ABC, abstractmethod
import warnings


class RankingAlgorithm(ABC):
    """
    This callable class implement an edge ranking algorithm used by the graph exploration algorithms.
    It is basically a function that takes a sample and generate neighbors; potentially similar
    and adversarial samples, and/or those samples that can be further used to explore more samples.
    A subclass should implement `neighbors`. The exploration algorithms use this as a callable,
    and `__call__` just invokes `neighbors`.

    Attributes
    ----------
    transformer_list: List of available transformations to be used by the algorithm.
    """

    def __init__(self, transformer_list, multi_feature_input=False, is_trained=True):
        """
        :param transformer_list: List of available transformations to be used by the algorithm. Each entry should be a
        transformers or [transformer_object, input_index] item
        :param multi_feature_input: A boolean flag indicating if each index in the transformer
        list defines a transformer(s) for a single data type in the input. If False, then it
        is assumed that the transformer list defines multiple transformers for a single input
        :param dependencies: A list of lists. Each element is the for [dependency_function, **dependency_args]. These functions
        enforce inter-feature constraints that much be maintained when a feature is modified.
        :param trained: A boolean value indicating if the ranking algorithm needs training. This is not necessary for every
        algorithm
        """

        if not isinstance(transformer_list, list):
            raise ValueError(f"Transformer list must be a list of transformers or [transformer, input_index] items")

        # Reformat transformer list if necessary
        if not isinstance(transformer_list[0], list):
            if multi_feature_input:
                raise ValueError(
                    f"For multi-feature inputs, the transformer list must be a list of [transformer, input_index] items"
                )
            self.transformer_list = []
            for transformer in transformer_list:
                self.transformer_list.append([transformer, None])
        else:
            self.transformer_list = transformer_list
        self.multi_feature_input = multi_feature_input

        self.is_trained = is_trained

    @abstractmethod
    def rank_edges(self, sample, scoring_function, score_input, dependencies=[], current_transformation_records=None):
        raise NotImplementedError

    def _train(self, training_samples, value_function, value_input_function, scoring_alg, dependencies=[], **kwargs):
        """
        Train the ranking algorithm. This is used by GraphExplorer since it holds some of the values we need for training
        :param training_samples: Input samples used to train the action dictionary
        :param value_function: Function used to compute values to compute the value a transformation has on the input with respect to a certain loss.
        :param value_input_function: Function used to compute the initial state of the sample before transform. This is fed into the value function as an input.
        :param scoring_alg: Indicates the scoring alg used for ranking if necessary
        :param dependencies: Input dependencies that need to be enforced
        """
        print(f"Training method not defined")

    def _enforce_dependencies(self, sample, dependencies):
        """
        Enforces dependencies specified in the list. Each element in dependencies should be of the form
        [callable_dependency function, dependency arguments]
        """

        if not self.multi_feature_input and len(dependencies) != 0:
            warnings.warn(
                f"Usually, dependencies are provided for multi-feature inputs. We'll keep going, but watch out!"
            )

        for d in dependencies:
            sample = d[0](sample, **d[1])
        return sample
